from .. import (
    EngineEnvironment,
    InputType,
    NoOperationAvailableException,
    AgentOperation,
)
from ..engine import EngineAgent
from ...entities import (
    EngineMessage,
    Input,
    Message,
    MessageRole,
)
from .response.orchestrator_response import OrchestratorResponse
from ..renderer import Renderer, TemplateEngineAgent
from ...event import Event, EventType
from ...event.manager import EventManager
from ...memory.manager import MemoryManager
from ...model.engine import Engine
from ...model.manager import ModelManager
from ...entities import Modality as Modality
from ...tool.manager import ToolManager
from contextlib import ExitStack
from dataclasses import asdict
from json import dumps
from logging import Logger
from time import perf_counter
from typing import Any
from uuid import UUID, uuid4


class Orchestrator:
    _id: UUID
    _name: str | None
    _operations: list[AgentOperation]
    _renderer: Renderer
    _total_operations: int
    _logger: Logger
    _model_manager: ModelManager
    _memory: MemoryManager
    _tool: ToolManager
    _event_manager: EventManager
    _engine_agents: dict[EngineEnvironment, EngineAgent] = {}
    _engines_stack: ExitStack = ExitStack()
    _operation_step: int | None = None
    _model_ids: set[str] = set()
    _call_options: dict | None = None
    _last_engine_agent: EngineAgent | None = None
    _exit_memory: bool = True

    def __init__(
        self,
        logger: Logger,
        model_manager: ModelManager,
        memory: MemoryManager,
        tool: ToolManager,
        event_manager: EventManager,
        operations: AgentOperation | list[AgentOperation],
        *,
        call_options: dict | None = None,
        exit_memory: bool = True,
        id: UUID | None = None,
        name: str | None = None,
        renderer: Renderer | None = None,
    ):
        self._logger = logger
        self._model_manager = model_manager
        self._memory = memory
        self._tool = tool
        self._event_manager = event_manager
        self._operations = (
            [operations]
            if isinstance(operations, AgentOperation)
            else operations
        )
        self._id = id or uuid4()
        self._exit_memory = exit_memory
        self._name = name
        self._renderer = renderer or Renderer()
        self._total_operations = len(self._operations)
        self._call_options = call_options

    @property
    def engine_agent(self) -> EngineAgent | None:
        return self._last_engine_agent

    @property
    def engine(self) -> Engine | None:
        return (
            self._last_engine_agent.engine if self._last_engine_agent else None
        )

    @property
    def id(self) -> UUID:
        return self._id

    @property
    def input_token_count(self) -> int | None:
        return (
            self._last_engine_agent.input_token_count
            if self._last_engine_agent
            else None
        )

    @property
    def is_finished(self) -> bool:
        return (
            self._operation_step is not None
            and self._operation_step == self._total_operations - 1
        )

    @property
    def memory(self) -> MemoryManager:
        return self._memory

    @property
    def model_ids(self) -> set[str]:
        return self._model_ids

    @property
    def name(self) -> str | None:
        return self._name

    @property
    def operations(self) -> list[AgentOperation]:
        return self._operations

    @property
    def tool(self) -> ToolManager:
        return self._tool

    @property
    def event_manager(self) -> EventManager:
        return self._event_manager

    @property
    def renderer(self) -> Renderer:
        """Return the renderer used by the orchestrator."""
        return self._renderer

    async def __call__(self, input: Input, **kwargs) -> OrchestratorResponse:
        tool_confirm = kwargs.pop("tool_confirm", None)
        if self.is_finished:
            self._operation_step = 0

        # Pick next operation step
        operation_step = (
            self._operation_step + 1
            if self._operation_step
            and self._operation_step < self._total_operations
            else 0 if not self._operation_step else None
        )
        self._operation_step = operation_step
        if self._operation_step is None:
            raise NoOperationAvailableException()

        # Load engine agent
        operation = self._operations[self._operation_step]
        environment_hash = dumps(asdict(operation.environment))
        assert self._engine_agents and environment_hash in self._engine_agents
        engine_agent = self._engine_agents[environment_hash]

        # Adapt tool manager
        if (
            engine_agent.engine.tokenizer
            and engine_agent.engine.tokenizer.eos_token
        ):
            self._tool.set_eos_token(engine_agent.engine.tokenizer.eos_token)

        await self._event_manager.trigger(
            Event(type=EventType.START, payload={"step": self._operation_step})
        )

        # Validate input
        input_type = operation.specification.input_type
        assert (
            input_type != InputType.TEXT
            or isinstance(input, str)
            or isinstance(input, Message)
            or isinstance(input, list)
        )

        if input_type == InputType.TEXT and isinstance(input, str):
            input = Message(role=MessageRole.USER, content=input)

        # Execute operation
        engine_args = {**(self._call_options or {}), **kwargs}
        start = perf_counter()
        await self._event_manager.trigger(
            Event(
                type=EventType.ENGINE_RUN_BEFORE,
                payload={
                    "input": input,
                    "specification": operation.specification,
                },
                started=start,
            )
        )
        result = await engine_agent(
            operation.specification, input, **engine_args
        )
        end = perf_counter()
        await self._event_manager.trigger(
            Event(
                type=EventType.ENGINE_RUN_AFTER,
                payload={
                    "result": result,
                    "input": input,
                    "specification": operation.specification,
                },
                started=start,
                finished=end,
                elapsed=end - start,
            )
        )

        self._last_engine_agent = engine_agent

        return OrchestratorResponse(
            input,
            result,
            engine_agent,
            operation,
            engine_args,
            event_manager=self._event_manager,
            tool=self._tool,
            tool_confirm=tool_confirm,
            agent_id=self._id,
            participant_id=self._memory.participant_id,
            session_id=(
                self._memory.permanent_message.session_id
                if self._memory.permanent_message
                else None
            ),
        )

    async def __aenter__(self):
        first_agent: TemplateEngineAgent | None = None
        model_ids: list[str] = []
        for operation in self._operations:
            # Load engine with environment
            environment = operation.environment
            environment_hash = dumps(asdict(environment))
            if environment_hash not in self._engine_agents:
                model_ids.append(environment.engine_uri.model_id)
                engine = self._model_manager.load_engine(
                    environment.engine_uri,
                    environment.settings,
                    operation.modality,
                )
                if not engine:
                    raise NotImplementedError()

                self._engines_stack.enter_context(engine)
                agent = TemplateEngineAgent(
                    engine,
                    self._memory,
                    self._tool,
                    self._event_manager,
                    self._model_manager,
                    self._renderer,
                    environment.engine_uri,
                    name=self._name,
                    id=self._id,
                )
                self._engine_agents[environment_hash] = agent
                if not first_agent:
                    first_agent = agent

        self._last_engine_agent = first_agent
        self._model_ids = set(model_ids)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any | None,
    ):
        if (
            self._last_engine_agent
            and self._last_engine_agent.output
            and (
                self._memory.has_permanent_message
                or self._memory.has_recent_message
            )
        ):
            previous_message = Message(
                role=MessageRole.ASSISTANT,
                content=await self._last_engine_agent.output.to_str(),
            )
            await self._memory.append_message(
                EngineMessage(
                    agent_id=self._id,
                    model_id=self._last_engine_agent.engine.model_id,
                    message=previous_message,
                )
            )

        if self._exit_memory:
            self._memory.__exit__(exc_type, exc_value, traceback)

        return self._engines_stack.__exit__(exc_type, exc_value, traceback)



from ...compat import override
from ...model.vendor import TextGenerationVendor
from ...model.audio import BaseAudioModel
from ...model.engine import Engine
from diffusers import DiffusionPipeline
from torch import from_numpy, inference_mode
from torchaudio import save
from transformers import (
    AutoProcessor,
    MusicgenForConditionalGeneration,
    PreTrainedModel,
)
from typing import Literal


class AudioGenerationModel(BaseAudioModel):
    _processor: AutoProcessor

    def _load_model(
        self,
    ) -> PreTrainedModel | TextGenerationVendor | DiffusionPipeline:
        self._processor = AutoProcessor.from_pretrained(self._model_id)
        model = MusicgenForConditionalGeneration.from_pretrained(
            self._model_id,
            device_map=self._device,
            tp_plan=Engine._get_tp_plan(self._settings.parallel),
            subfolder=self._settings.subfolder or "",
        ).to(self._device)
        return model

    @override
    async def __call__(
        self,
        prompt: str,
        path: str,
        max_new_tokens: int,
        *,
        padding: bool = True,
        tensor_format: Literal["pt"] = "pt",
    ) -> str:
        assert path

        inputs = self._processor(
            text=[prompt], return_tensors=tensor_format, padding=padding
        )
        inputs.to(self._device)

        with inference_mode():
            audio_tokens = self._model.generate(
                **inputs, max_new_tokens=max_new_tokens
            )

        sampling_rate = self._model.config.audio_encoder.sampling_rate
        waveform = audio_tokens[0, 0].cpu().numpy()
        wave_tensor = from_numpy(waveform).unsqueeze(0)
        save(path, wave_tensor, sampling_rate)

        return path



from ....vendor import TextGenerationVendor, TextGenerationVendorStream
from . import TextGenerationVendorModel
from .....compat import override
from .....entities import GenerationSettings, Message, Token, TokenDetail
from .....tool.manager import ToolManager
from diffusers import DiffusionPipeline
from transformers import PreTrainedModel
from typing import AsyncIterator
import litellm


class LiteLLMStream(TextGenerationVendorStream):
    def __init__(self, stream: AsyncIterator):
        super().__init__(stream.__aiter__())

    async def __anext__(self) -> Token | TokenDetail | str:
        chunk = await self._generator.__anext__()
        choice = None
        if isinstance(chunk, dict):
            choice = chunk.get("choices", [{}])[0]
            delta = choice.get("delta", {}) if isinstance(choice, dict) else {}
            text = delta.get("content", "")
        else:
            choice = chunk.choices[0]
            delta = getattr(choice, "delta", None)
            text = getattr(delta, "content", "") if delta else ""
        return text


class LiteLLMClient(TextGenerationVendor):
    _api_key: str | None
    _base_url: str | None

    def __init__(
        self, api_key: str | None = None, base_url: str | None = None
    ):
        self._api_key = api_key
        self._base_url = base_url or "http://localhost:4000"

    @override
    async def __call__(
        self,
        model_id: str,
        messages: list[Message],
        settings: GenerationSettings | None = None,
        *,
        tool: ToolManager | None = None,
        use_async_generator: bool = True,
    ) -> AsyncIterator[Token | TokenDetail | str]:
        template_messages = self._template_messages(messages)
        kwargs = dict(
            model=model_id,
            messages=template_messages,
            api_key=self._api_key,
            stream=use_async_generator,
        )
        if self._base_url:
            kwargs["api_base"] = self._base_url
        result = await litellm.acompletion(**kwargs)
        if use_async_generator:
            return LiteLLMStream(result)

        async def single_gen():
            if isinstance(result, dict):
                yield result["choices"][0]["message"]["content"]
            else:
                yield result.choices[0].message.content or ""

        return single_gen()


class LiteLLMModel(TextGenerationVendorModel):
    def _load_model(
        self,
    ) -> PreTrainedModel | TextGenerationVendor | DiffusionPipeline:
        return LiteLLMClient(
            api_key=self._settings.access_token,
            base_url=self._settings.base_url,
        )



from ..entities import EngineMessage
from ..event import Event, EventType
from ..event.manager import EventManager
from ..memory import RecentMessageMemory
from ..memory.partitioner.text import TextPartitioner
from ..memory.permanent import (
    PermanentMessageMemory,
    PermanentMemory,
    VectorFunction,
)
from time import perf_counter
from logging import Logger
from typing import Any
from uuid import UUID


class MemoryManager:
    _agent_id: UUID
    _participant_id: UUID
    _permanent_message_memory: PermanentMessageMemory | None = None
    _permanent_memories: dict[str, PermanentMemory]
    _recent_message_memory: RecentMessageMemory | None = None
    _text_partitioner: TextPartitioner
    _logger: Logger
    _event_manager: EventManager | None = None

    @classmethod
    async def create_instance(
        cls,
        *args,
        agent_id: UUID,
        participant_id: UUID,
        text_partitioner: TextPartitioner,
        logger: Logger,
        with_permanent_message_memory: str | None = None,
        with_recent_message_memory: bool = True,
        event_manager: EventManager | None = None,
    ):
        permanent_memory: PermanentMessageMemory | None = None
        if with_permanent_message_memory:
            from .permanent.pgsql.message import PgsqlMessageMemory

            permanent_memory = await PgsqlMessageMemory.create_instance(
                dsn=with_permanent_message_memory,
                logger=logger,
            )
        recent_memory = (
            RecentMessageMemory() if with_recent_message_memory else None
        )

        manager = cls(
            agent_id=agent_id,
            participant_id=participant_id,
            permanent_message_memory=permanent_memory,
            recent_message_memory=recent_memory,
            text_partitioner=text_partitioner,
            logger=logger,
            event_manager=event_manager,
        )
        return manager

    def __init__(
        self,
        *args,
        agent_id: UUID,
        participant_id: UUID,
        permanent_message_memory: PermanentMessageMemory | None,
        recent_message_memory: RecentMessageMemory | None,
        text_partitioner: TextPartitioner,
        logger: Logger,
        event_manager: EventManager | None = None,
        permanent_memories: dict[str, PermanentMemory] | None = None,
    ):
        assert agent_id and participant_id
        self._logger = logger
        self._agent_id = agent_id
        self._participant_id = participant_id
        self._text_partitioner = text_partitioner
        self._permanent_memories = {}
        self._event_manager = event_manager
        if permanent_message_memory:
            self.add_permanent_message_memory(permanent_message_memory)
        if recent_message_memory:
            self.add_recent_message_memory(recent_message_memory)
        if permanent_memories:
            for namespace, memory in permanent_memories.items():
                self.add_permanent_memory(namespace, memory)

    @property
    def participant_id(self) -> UUID:
        """Return the participant identifier associated with this memory."""
        return self._participant_id

    @property
    def has_permanent_message(self) -> bool:
        return bool(self._permanent_message_memory)

    @property
    def has_recent_message(self) -> bool:
        return bool(self._recent_message_memory)

    @property
    def permanent_message(self) -> PermanentMessageMemory | None:
        return self._permanent_message_memory

    @property
    def recent_message(self) -> RecentMessageMemory | None:
        return self._recent_message_memory

    @property
    def recent_messages(self) -> list[EngineMessage] | None:
        return (
            self._recent_message_memory.data
            if self._recent_message_memory
            else None
        )

    def add_recent_message_memory(self, memory: RecentMessageMemory):
        self._recent_message_memory = memory

    def add_permanent_message_memory(self, memory: PermanentMessageMemory):
        self._permanent_message_memory = memory

    def add_permanent_memory(
        self, namespace: str, memory: PermanentMemory
    ) -> None:
        assert namespace and memory
        self._permanent_memories[namespace] = memory

    def delete_permanent_memory(self, namespace: str) -> None:
        self._permanent_memories.pop(namespace, None)

    async def append_message(self, engine_message: EngineMessage) -> None:
        assert (
            isinstance(engine_message, EngineMessage)
            and engine_message.agent_id
            and engine_message.message
            and engine_message.message.content
        )

        self._logger.debug("<Memory> Appending message")

        if self._permanent_message_memory:
            start = perf_counter()
            if self._event_manager:
                await self._event_manager.trigger(
                    Event(
                        type=EventType.MEMORY_PERMANENT_MESSAGE_ADD,
                        payload={
                            "message": engine_message,
                            "participant_id": self._participant_id,
                            "session_id": (
                                self._permanent_message_memory.session_id
                                if self._permanent_message_memory
                                else None
                            ),
                        },
                        started=start,
                    )
                )
            partitions = await self._text_partitioner(
                engine_message.message.content
            )
            await self._permanent_message_memory.append_with_partitions(
                engine_message, partitions=partitions
            )
            if self._event_manager:
                end = perf_counter()
                await self._event_manager.trigger(
                    Event(
                        type=EventType.MEMORY_PERMANENT_MESSAGE_ADDED,
                        payload={
                            "message": engine_message,
                            "participant_id": self._participant_id,
                            "session_id": (
                                self._permanent_message_memory.session_id
                                if self._permanent_message_memory
                                else None
                            ),
                        },
                        started=start,
                        finished=end,
                        elapsed=end - start,
                    )
                )

        if self._recent_message_memory:
            self._recent_message_memory.append(engine_message)

        self._logger.debug("<Memory> Message appended")

    async def continue_session(
        self,
        session_id: UUID,
        *args,
        load_recent_messages: bool = True,
        load_recent_messages_limit: int | None = None,
    ) -> None:
        self._logger.debug("Continuing session %s", session_id)
        if self._permanent_message_memory:
            start = perf_counter()
            if self._event_manager:
                await self._event_manager.trigger(
                    Event(
                        type=EventType.MEMORY_PERMANENT_MESSAGE_SESSION_CONTINUE,
                        payload={
                            "session_id": session_id,
                            "participant_id": self._participant_id,
                        },
                        started=start,
                    )
                )
            await self._permanent_message_memory.continue_session(
                agent_id=self._agent_id,
                participant_id=self._participant_id,
                session_id=session_id,
            )

        if (
            load_recent_messages
            and self._permanent_message_memory
            and self._recent_message_memory
        ):
            messages = (
                await self._permanent_message_memory.get_recent_messages(
                    participant_id=self._participant_id,
                    session_id=session_id,
                    limit=load_recent_messages_limit,
                )
            )
            self._recent_message_memory.reset()
            for message in messages:
                self._recent_message_memory.append(message)

        self._logger.debug("Session %s continued", session_id)
        if self._permanent_message_memory and self._event_manager:
            end = perf_counter()
            await self._event_manager.trigger(
                Event(
                    type=EventType.MEMORY_PERMANENT_MESSAGE_SESSION_CONTINUED,
                    payload={
                        "session_id": session_id,
                        "participant_id": self._participant_id,
                    },
                    started=start,
                    finished=end,
                    elapsed=end - start,
                )
            )

    async def start_session(self) -> None:
        self._logger.debug("Starting session")
        if self._permanent_message_memory:
            start = perf_counter()
            if self._event_manager:
                await self._event_manager.trigger(
                    Event(
                        type=EventType.MEMORY_PERMANENT_MESSAGE_SESSION_START,
                        payload={"participant_id": self._participant_id},
                        started=start,
                    )
                )
            await self._permanent_message_memory.reset_session(
                agent_id=self._agent_id, participant_id=self._participant_id
            )

        if self._recent_message_memory:
            self._recent_message_memory.reset()

        self._logger.debug("Session started")
        if self._permanent_message_memory and self._event_manager:
            end = perf_counter()
            await self._event_manager.trigger(
                Event(
                    type=EventType.MEMORY_PERMANENT_MESSAGE_SESSION_STARTED,
                    payload={
                        "session_id": (
                            self._permanent_message_memory.session_id
                        ),
                        "participant_id": self._participant_id,
                    },
                    started=start,
                    finished=end,
                    elapsed=end - start,
                )
            )

    async def search_messages(
        self,
        search: str,
        agent_id: UUID,
        participant_id: UUID,
        *args,
        function: VectorFunction,
        limit: int | None = None,
        search_user_messages: bool = False,
        session_id: UUID | None = None,
        exclude_session_id: UUID | None = None,
    ) -> list[EngineMessage]:
        assert self._permanent_message_memory
        search_partitions = await self._text_partitioner(search)
        messages = await self._permanent_message_memory.search_messages(
            search_partitions=search_partitions,
            search_user_messages=search_user_messages,
            agent_id=agent_id,
            participant_id=participant_id,
            function=function,
            limit=limit,
            session_id=session_id,
            exclude_session_id=exclude_session_id,
        )
        return messages

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any | None,
    ):
        pass



from datetime import datetime, timezone
from logging import Logger
from typing import Any
from uuid import UUID, uuid4

from elasticsearch import AsyncElasticsearch

from ....entities import (
    EngineMessage,
    EngineMessageScored,
    Message,
    MessageRole,
)
from ....memory.partitioner.text import TextPartition
from ....memory.permanent import (
    PermanentMessageMemory,
    VectorFunction,
)
from . import ElasticsearchMemory


class ElasticsearchMessageMemory(ElasticsearchMemory, PermanentMessageMemory):
    def __init__(
        self,
        index: str,
        *,
        client: Any,
        logger: Logger,
        sentence_model: Any | None = None,
    ) -> None:
        ElasticsearchMemory.__init__(
            self, index=index, client=client, logger=logger
        )
        PermanentMessageMemory.__init__(self, sentence_model=sentence_model)  # type: ignore[arg-type]

    @classmethod
    async def create_instance(
        cls,
        index: str,
        *,
        logger: Logger,
        es_client: Any | None = None,
        sentence_model: Any | None = None,
    ) -> "ElasticsearchMessageMemory":
        if es_client is None:
            es_client = AsyncElasticsearch()
        return cls(
            index=index,
            client=es_client,
            logger=logger,
            sentence_model=sentence_model,
        )

    async def create_session(
        self, *, agent_id: UUID, participant_id: UUID
    ) -> UUID:
        now_utc = datetime.now(timezone.utc)
        session = self._build_session(
            agent_id,
            participant_id,
            created_at=now_utc,
        )
        await self._index_document(
            index=self._index,
            id=f"{session.id}.json",
            document={
                "id": str(session.id),
                "agent_id": str(session.agent_id),
                "participant_id": str(session.participant_id),
                "messages": session.messages,
                "created_at": session.created_at.isoformat(),
            },
        )
        return session.id

    async def continue_session_and_get_id(
        self,
        *,
        agent_id: UUID,
        participant_id: UUID,
        session_id: UUID,
    ) -> UUID:
        result = await self._get_document(
            index=self._index, id=f"{session_id}.json"
        )
        meta = result.get("_source") if result else None
        assert meta
        assert meta.get("agent_id") == str(agent_id)
        assert meta.get("participant_id") == str(participant_id)
        return session_id

    async def append_with_partitions(
        self,
        engine_message: EngineMessage,
        *,
        partitions: list[TextPartition],
    ) -> None:
        assert engine_message and partitions
        now_utc = datetime.now(timezone.utc)
        message, message_partitions = self._build_message_with_partitions(
            engine_message,
            self._session_id,
            partitions,
            created_at=now_utc,
            message_id=uuid4(),
        )
        key = (
            f"{self._index}/{message.session_id}/{message.id}.json"
            if message.session_id
            else f"{self._index}/{message.id}.json"
        )
        await self._index_document(
            index=self._index,
            id=key,
            document={
                "id": str(message.id),
                "agent_id": str(message.agent_id),
                "model_id": message.model_id,
                "session_id": (
                    str(message.session_id) if message.session_id else None
                ),
                "author": str(message.author),
                "data": message.data,
                "partitions": message.partitions,
                "created_at": message.created_at.isoformat(),
            },
        )
        for row in message_partitions:
            await self._index_vector(
                index=self._index,
                id=f"{row.message_id}:{row.partition}",
                vector=row.embedding.tolist(),
                metadata={
                    "message_id": str(row.message_id),
                    "agent_id": str(row.agent_id),
                    "session_id": (
                        str(row.session_id) if row.session_id else None
                    ),
                },
            )

    async def get_recent_messages(
        self,
        session_id: UUID,
        participant_id: UUID,
        *,
        limit: int | None = None,
    ) -> list[EngineMessage]:
        response = await self._call_client(
            self._client.search,
            index=self._index,
            body={
                "query": {"term": {"session_id": str(session_id)}},
                "sort": [{"created_at": {"order": "desc"}}],
                "size": limit or 10,
            },
        )
        hits = response.get("hits", {}).get("hits", []) if response else []
        messages: list[EngineMessage] = []
        for hit in hits:
            meta = hit.get("_source")
            if not meta:
                continue
            messages.append(
                EngineMessage(
                    agent_id=UUID(meta["agent_id"]),
                    model_id=meta["model_id"],
                    message=Message(
                        role=MessageRole(meta["author"]),
                        content=meta["data"],
                    ),
                )
            )
        return messages

    async def search_messages(
        self,
        *,
        agent_id: UUID,
        function: VectorFunction,
        limit: int | None = None,
        participant_id: UUID,
        search_partitions: list[TextPartition],
        search_user_messages: bool,
        session_id: UUID | None,
        exclude_session_id: UUID | None,
    ) -> list[EngineMessageScored]:
        assert agent_id and participant_id and search_partitions
        query = search_partitions[0].embeddings.tolist()
        filt = {
            "message_id": "*",
            "agent_id": str(agent_id),
            "participant_id": str(participant_id),
        }
        if session_id:
            filt["session_id"] = str(session_id)
        response = await self._query_vector(
            index=self._index,
            query_vector=query,
            top_k=limit or 10,
            function=str(function),
            filter=filt,
        )
        results: list[EngineMessageScored] = []
        for item in response.get("Items", []):
            msg_id = item.get("Metadata", {}).get("message_id")
            if not msg_id:
                continue
            key = (
                f"{self._index}/{session_id}/{msg_id}.json"
                if session_id
                else f"{self._index}/{msg_id}.json"
            )
            obj = await self._get_document(index=self._index, id=key)
            meta = obj.get("_source") if obj else None
            if not meta:
                continue
            results.append(
                EngineMessageScored(
                    agent_id=UUID(meta["agent_id"]),
                    model_id=meta["model_id"],
                    message=Message(
                        role=MessageRole(meta["author"]), content=meta["data"]
                    ),
                    score=item.get("Score", 0.0),
                )
            )
        return results



from ...agent.orchestrator import Orchestrator
from ...cli import confirm, get_input, has_input
from ...cli.commands.cache import cache_delete, cache_download
from ...entities import (
    Model,
    Modality,
    ReasoningToken,
    Token,
    ToolCallToken,
)
from ...entities import GenerationSettings  # noqa: F401
from ...event import Event, EventStats, EventType, TOOL_TYPES
from ...model.criteria import KeywordStoppingCriteria  # noqa: F401
from ...model.hubs.huggingface import HuggingfaceHub
from ...model.manager import ModelManager
from ...model.nlp.sentence import SentenceTransformerModel
from ...model.nlp.text.generation import TextGenerationModel
from ...model.response.text import TextGenerationResponse
from ...secrets import KeyringSecrets
from . import get_model_settings
from dataclasses import replace
from argparse import Namespace
from asyncio import (
    as_completed,
    create_task,
    gather,
    sleep,
    to_thread,
    Event as EventSignal,
)
from datetime import datetime, timezone
from logging import Logger
from rich.console import Console, Group, RenderableType
from rich.live import Live
from rich.padding import Padding
from rich.prompt import Prompt
from rich.spinner import Spinner
from rich.theme import Theme
from time import perf_counter


def model_display(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
    *vargs,
    modality: Modality | None = None,
    load: bool | None = None,
    model: SentenceTransformerModel | TextGenerationModel | None = None,
    summary: bool | None = None,
) -> None:
    assert args.model
    _ = theme._

    with ModelManager(hub, logger) as manager:
        engine_uri = manager.parse_uri(args.model)
        model_id = args.model
        can_access = args.skip_hub_access_check or hub.can_access(model_id)
        hub_model = hub.model(model_id)
        console.print(
            theme.model(
                hub_model,
                can_access=can_access,
                expand=(summary is not None and not summary)
                or (summary is None and not args.summary),
                summary=False,
            )
        )

        is_runnable = not engine_uri.is_local
        if not model and (
            (load is not None and load) or (load is None and args.load)
        ):
            model_settings = get_model_settings(
                args,
                hub,
                logger,
                engine_uri,
                modality=modality,
            )
            with manager.load(**model_settings) as lm:
                logger.debug("Loaded model %s", lm.config.__repr__())
                is_runnable = lm.is_runnable(getattr(args, "device", None))
                console.print(
                    Padding(
                        theme.model_display(
                            lm.config,
                            lm.tokenizer_config,
                            is_runnable=is_runnable,
                            summary=summary or False,
                        ),
                        pad=(0, 0, 0, 0),
                    )
                )
        elif model:
            console.print(
                Padding(
                    theme.model_display(
                        model.config,
                        model.tokenizer_config,
                        is_runnable=is_runnable,
                        summary=summary or False,
                    ),
                    pad=(0, 0, 0, 0),
                )
            )


def model_install(
    args: Namespace, console: Console, theme: Theme, hub: HuggingfaceHub
) -> None:
    assert args.model
    engine_uri = ModelManager.parse_uri(args.model)
    if (
        engine_uri.vendor
        and engine_uri.password
        and engine_uri.user == "secret"
    ):
        secrets = KeyringSecrets()
        token = secrets.read(engine_uri.password)
        if token is None:
            secret_value = Prompt.ask(
                theme.ask_secret_password(engine_uri.password)
            )
            secrets.write(engine_uri.password, secret_value)
        elif confirm(console, theme.ask_override_secret(engine_uri.password)):
            secret_value = Prompt.ask(
                theme.ask_secret_password(engine_uri.password)
            )
            secrets.write(engine_uri.password, secret_value)

    cache_download(args, console, theme, hub)


async def model_run(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    refresh_per_second: int,
    logger: Logger,
) -> None:
    assert args.model and args.device and args.max_new_tokens
    _, _i = theme._, theme.icons

    with ModelManager(hub, logger) as manager:
        engine_uri = manager.parse_uri(args.model)
        model_settings = get_model_settings(args, hub, logger, engine_uri)
        modality = model_settings["modality"]
        assert modality

        if not args.quiet:
            if engine_uri.is_local:
                can_access = (
                    args.quiet
                    or args.skip_hub_access_check
                    or hub.can_access(engine_uri.model_id)
                )

                model = hub.model(engine_uri.model_id)
                console.print(
                    Padding(
                        theme.model(
                            model, can_access=can_access, summary=True
                        ),
                        pad=(0, 0, 1, 0),
                    )
                )

        operation = ModelManager.get_operation_from_arguments(
            modality, args, None
        )

        with manager.load(**model_settings) as model:
            logger.debug("Loaded model %s", model.config.__repr__())

            if operation.requires_input or has_input(console):
                input_string = get_input(
                    console,
                    _i["user_input"] + " ",
                    echo_stdin=not args.no_repl,
                    is_quiet=args.quiet,
                )
                if not input_string:
                    return

                operation = replace(operation, input=input_string)

            output = await manager(engine_uri, model, operation)

            if operation.modality in {
                Modality.AUDIO_SPEECH_RECOGNITION,
                Modality.TEXT_QUESTION_ANSWERING,
                Modality.TEXT_SEQUENCE_CLASSIFICATION,
                Modality.TEXT_SEQUENCE_TO_SEQUENCE,
                Modality.TEXT_TRANSLATION,
                Modality.VISION_IMAGE_TO_TEXT,
                Modality.VISION_ENCODER_DECODER,
                Modality.VISION_IMAGE_TEXT_TO_TEXT,
            }:
                console.print(output)

            elif operation.modality == Modality.AUDIO_CLASSIFICATION:
                console.print(theme.display_audio_labels(output))

            elif operation.modality == Modality.AUDIO_TEXT_TO_SPEECH:
                console.print(f"Audio generated in {output}")

            elif operation.modality == Modality.AUDIO_GENERATION:
                console.print(f"Audio generated in {output}")

            elif operation.modality == Modality.TEXT_TOKEN_CLASSIFICATION:
                console.print(theme.display_token_labels([output]))

            elif operation.modality == Modality.TEXT_GENERATION:
                await token_generation(
                    args=args,
                    console=console,
                    theme=theme,
                    logger=logger,
                    orchestrator=None,
                    event_stats=None,
                    lm=model,
                    input_string=operation.input,
                    refresh_per_second=refresh_per_second,
                    response=output,
                    dtokens_pick=operation.parameters["text"].pick_tokens,
                    display_tokens=args.display_tokens or 0,
                    with_stats=not args.quiet,
                    tool_events_limit=args.display_tools_events,
                )

            elif operation.modality == Modality.VISION_IMAGE_CLASSIFICATION:
                console.print(theme.display_image_entity(output))

            elif operation.modality == Modality.VISION_OBJECT_DETECTION:
                console.print(theme.display_image_entities(output, sort=True))

            elif operation.modality == Modality.VISION_SEMANTIC_SEGMENTATION:
                console.print(theme.display_image_labels(output))

            elif operation.modality == Modality.VISION_TEXT_TO_IMAGE:
                console.print(output)

            elif operation.modality == Modality.VISION_TEXT_TO_ANIMATION:
                console.print(output)

            elif operation.modality == Modality.VISION_TEXT_TO_VIDEO:
                console.print(output)

            else:
                raise NotImplementedError(
                    f"Modality {operation.modality} not supported"
                )


async def model_search(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    refresh_per_second: int,
) -> None:
    assert args.limit
    _ = theme._

    models: list[Model] = []
    model_access: dict[str, bool] = {}

    # Fetch matching models
    with console.status(
        _("Loading models..."),
        spinner=theme.get_spinner("downloading"),
        refresh_per_second=refresh_per_second,
    ):
        models = [
            model
            for model in hub.models(
                filter=args.filter or None,
                search=args.search or None,
                library=args.library or None,
                author=args.author,
                gated=True if args.gated else False if args.open else None,
                language=args.language or None,
                name=args.name or None,
                task=args.task or None,
                tags=args.tag or None,
                limit=args.limit,
            )
        ]

    # Tasks to check model access
    tasks = [
        create_task(to_thread(lambda id=model.id: (id, hub.can_access(id))))
        for model in models
    ]

    def _render(
        models: list[Model], model_access: dict[str, bool]
    ) -> list[RenderableType]:
        return [
            theme.model(
                model,
                can_access=(
                    model_access[model.id]
                    if model.id in model_access
                    else None
                ),
            )
            for model in models
        ]

    # Keep list of models updated as tasks are completed
    with Live(
        Group(*_render(models, model_access)),
        console=console,
        refresh_per_second=refresh_per_second,
    ) as live:
        for completed_task in as_completed(tasks):
            model_id, can_access = await completed_task
            model_access[model_id] = can_access

            live.update(Group(*_render(models, model_access)))

        await gather(*tasks)


def model_uninstall(
    args: Namespace, console: Console, theme: Theme, hub: HuggingfaceHub
) -> None:
    assert args.model
    engine_uri = ModelManager.parse_uri(args.model)
    if (
        engine_uri.vendor
        and engine_uri.password
        and engine_uri.user == "secret"
    ):
        secrets = KeyringSecrets()
        secrets.delete(engine_uri.password)

    cache_delete(args, console, theme, hub, is_full_deletion=True)


async def token_generation(
    args: Namespace,
    console: Console,
    theme: Theme,
    logger: Logger,
    orchestrator: Orchestrator | None,
    event_stats: EventStats | None,
    lm: TextGenerationModel,
    input_string: str,
    response: TextGenerationResponse,
    *,
    display_tokens: int,
    dtokens_pick: int,
    refresh_per_second: int,
    tool_events_limit: int | None,
    with_stats: bool = True,
):
    # If no statistics needed, return as early as possible
    if not with_stats:
        async for token in response:
            if isinstance(token, Event):
                continue
            text_token = token.token if isinstance(token, Token) else token
            console.print(text_token, end="")
        return

    # From here on, display includes stats and may include token probabilities

    if not orchestrator or (
        not args.display_events and not args.display_tools
    ):
        with Live(
            refresh_per_second=refresh_per_second, screen=args.record
        ) as live:
            await _token_stream(
                args,
                console,
                live,
                None,
                None,
                theme,
                logger,
                orchestrator,
                event_stats,
                lm,
                input_string,
                response,
                display_tokens=display_tokens,
                dtokens_pick=dtokens_pick,
                refresh_per_second=refresh_per_second,
                stop_signal=None,
                tool_events_limit=tool_events_limit,
                with_stats=with_stats,
            )
        return

    stop_signal = EventSignal()
    events_height = 6
    tools_height = 10
    empty = ""
    group = Group(empty, empty, empty)
    events_group_index = 0
    tools_group_index = 1
    tokens_group_index = 2

    with Live(
        group, refresh_per_second=refresh_per_second, screen=args.record
    ) as live:
        await gather(
            _event_stream(
                args,
                console,
                live,
                group,
                events_group_index,
                tools_group_index,
                orchestrator,
                theme,
                events_height=events_height,
                tools_height=tools_height,
                stop_signal=stop_signal,
            ),
            _token_stream(
                args,
                console,
                live,
                group,
                tokens_group_index,
                theme,
                logger,
                orchestrator,
                event_stats,
                lm,
                input_string,
                response,
                display_tokens=display_tokens,
                dtokens_pick=dtokens_pick,
                refresh_per_second=refresh_per_second,
                stop_signal=stop_signal,
                tool_events_limit=tool_events_limit,
                with_stats=with_stats,
            ),
        )


async def _event_stream(
    args: Namespace,
    console: Console,
    live: Live,
    group: Group,
    events_group_index: int,
    tools_group_index: int,
    orchestrator: Orchestrator,
    theme: Theme,
    *,
    events_height: int = 6,
    tools_height: int = 10,
    stop_signal: EventSignal,
) -> None:
    event_manager = orchestrator.event_manager
    if not event_manager or (
        not args.display_events and not args.display_tools
    ):
        return

    async for e in event_manager.listen(stop_signal=stop_signal):
        tool_view = e.type in TOOL_TYPES
        if (tool_view and not args.display_tools) or (
            not tool_view and not args.display_events
        ):
            continue

        events_renderable = theme.events(
            event_manager.history,
            events_limit=6 if tool_view else 4,
            height=tools_height if tool_view else events_height,
            include_tokens=False,
            include_tools=tool_view,
            include_tool_detect=False,
            include_non_tools=not tool_view,
            tool_view=tool_view,
        )
        if not events_renderable:
            continue

        _render_frame(
            args,
            console,
            live,
            events_renderable,
            group,
            tools_group_index if tool_view else events_group_index,
        )


async def _token_stream(
    args: Namespace,
    console: Console,
    live: Live,
    group: Group | None,
    tokens_group_index: int | None,
    theme: Theme,
    logger: Logger,
    orchestrator: Orchestrator | None,
    event_stats: EventStats | None,
    lm: TextGenerationModel,
    input_string: str,
    response: TextGenerationResponse,
    *,
    display_tokens: int,
    dtokens_pick: int,
    refresh_per_second: int,
    stop_signal: EventSignal | None,
    tool_events_limit: int | None,
    with_stats: bool = True,
) -> None:
    display_time_to_n_token = args.display_time_to_n_token
    display_reasoning_time = not getattr(args, "skip_display_reasoning_time")
    display_pause = (
        args.display_pause
        if args.display_pause and args.display_pause > 0
        else 0
    )
    start_thinking = (
        args.start_thinking if hasattr(args, "start_thinking") else False
    )
    tokens = []
    answer_text_tokens: list[str] = []
    thinking_text_tokens: list[str] = []
    tool_text_tokens: list[str] = []
    tool_events: list[Event] = []
    tool_event_calls: list[Event] = []
    tool_event_results: list[Event] = []
    completed_call_ids: set[str] = set()
    total_tokens = 0
    frame_minimum_pause_ms = (
        100 if display_pause > 0 and display_tokens > 0 else 0
    )

    input_token_count = (
        response.input_token_count
        if response.input_token_count
        else (
            orchestrator.input_token_count
            if orchestrator
            else lm.input_token_count(input_string)
        )
    )
    ttft: float | None = None
    ttnt: float | None = None
    last_current_dtoken: Token | None = None
    tool_running_spinner: Spinner | None = None

    if start_thinking and response.can_think and not response.is_thinking:
        response.set_thinking(start_thinking)

    start = perf_counter()
    started_reasoning = perf_counter() if response.is_thinking else None
    reasoning_time = None

    async for token in response:
        is_reasoning_token = isinstance(token, ReasoningToken)

        if isinstance(token, Event):
            event = token
            tool_events.append(event)
            if event.type == EventType.TOOL_MODEL_RESPONSE:
                tokens = []
                answer_text_tokens = []
                tool_text_tokens = []
                thinking_text_tokens = []
                inner_response = event.payload["response"]
                assert isinstance(inner_response, TextGenerationResponse)
                if inner_response.input_token_count:
                    input_token_count = inner_response.input_token_count
            elif event.type == EventType.TOOL_RESULT:
                tool_event_results.append(event)
                if "call" in event.payload:
                    completed_call_ids.add(event.payload["call"].id)
            else:
                tool_event_calls.append(event)
        else:
            if (
                display_reasoning_time
                and not is_reasoning_token
                and started_reasoning is not None
            ):
                reasoning_time = perf_counter() - started_reasoning
                started_reasoning = None

            text_token = token.token if isinstance(token, Token) else token
            if isinstance(token, ToolCallToken):
                tool_text_tokens.append(text_token)
            elif is_reasoning_token:
                if not started_reasoning:
                    started_reasoning = perf_counter()
                thinking_text_tokens.append(text_token)
            else:
                answer_text_tokens.append(text_token)

        tool_running_spinner = None
        if tool_event_calls or tool_event_results:
            tool_calling_names = [
                c.name
                for e in tool_event_calls
                for c in e.payload
                if c.id not in completed_call_ids
            ]

            tool_running_spinner = (
                Spinner(
                    theme.get_spinner("tool_running"),
                    text="[cyan]"
                    + theme._n(
                        "Running tool {tool_names}...",
                        "Running tools {tool_names}...",
                        len(tool_calling_names),
                    ).format(tool_names=", ".join(tool_calling_names))
                    + "[/cyan]",
                    style="cyan",
                    speed=1.0,
                )
                if tool_calling_names
                else None
            )

        elapsed = perf_counter() - start
        total_tokens += 1

        if ttft is None:
            ttft = elapsed
        if (
            ttnt is None
            and display_time_to_n_token
            and total_tokens >= display_time_to_n_token
        ):
            ttnt = elapsed

        ttsr = None
        if display_reasoning_time and reasoning_time:
            ttsr = reasoning_time

        if display_tokens and isinstance(token, Token):
            tokens.append(token)

        token_frames_promise = theme.tokens(
            lm.model_id,
            lm.tokenizer_config.tokens if lm.tokenizer_config else None,
            (
                lm.tokenizer_config.special_tokens
                if lm.tokenizer_config
                else None
            ),
            display_tokens,
            args.display_probabilities if dtokens_pick > 0 else False,
            dtokens_pick,
            # Which tokens to mark as interesting
            lambda dtoken: (
                (
                    dtoken.probability < args.display_probabilities_maximum
                    or len(
                        [
                            t
                            for t in dtoken.tokens
                            if t.id != dtoken.id
                            and t.probability
                            >= args.display_probabilities_sample_minimum
                        ]
                    )
                    > 0
                )
                if display_tokens
                and args.display_probabilities
                and args.display_probabilities_maximum > 0
                and args.display_probabilities_maximum > 0
                else None
            ),
            thinking_text_tokens,
            tool_text_tokens,
            answer_text_tokens,
            tokens or None,
            input_token_count,
            total_tokens,
            tool_events,
            tool_event_calls,
            tool_event_results,
            tool_running_spinner,
            ttft,
            ttnt,
            ttsr,
            elapsed,
            console.width,
            logger,
            event_stats,
            tool_events_limit=tool_events_limit,
            limit_answer_height=True,
            maximum_frames=1,
            start_thinking=start_thinking,
        )

        token_frame_list = [
            token_frame async for token_frame in token_frames_promise
        ]

        token_frames = [token_frame_list[0]]

        for current_dtoken, frame in token_frames:
            _render_frame(
                args, console, live, frame, group, tokens_group_index
            )

            if current_dtoken and current_dtoken != last_current_dtoken:
                last_current_dtoken = current_dtoken
                if display_pause > 0:
                    await sleep(display_pause / 1000)
                elif frame_minimum_pause_ms > 0:
                    await sleep(
                        frame_minimum_pause_ms / 1000
                    )  # pragma: no cover - unreachable
            elif (
                dtokens_pick > 0
                and not args.display_probabilities
                and display_pause > 0
            ):
                await sleep(display_pause / 1000)

        if (
            dtokens_pick > 0
            and args.display_probabilities
            and token_frame_list
            and len(token_frame_list) > 0
        ):
            for current_dtoken, frame in token_frame_list[1:]:
                _render_frame(
                    args, console, live, frame, group, tokens_group_index
                )

                if current_dtoken and display_pause > 0:
                    await sleep(display_pause / 1000)
                elif frame_minimum_pause_ms > 0:
                    await sleep(frame_minimum_pause_ms / 1000)

    if stop_signal:
        stop_signal.set()


def _render_frame(
    args: Namespace,
    console: Console,
    live: Live,
    frame: RenderableType,
    group: Group | None = None,
    group_index: int | None = None,
) -> None:
    if group and group_index is not None:
        group.renderables[group_index] = frame
        live.refresh()
    else:
        live.update(frame)

    if args.record:
        now = datetime.now(timezone.utc)
        ts = now.strftime("%Y%m%d%H%M%S")
        ms = now.microsecond // 1000
        filename = f"avalan-screenshot-{ts}-{ms:03d}.svg"
        console.save_svg(filename, clear=True)



from ....entities import EngineMessage, EngineMessageScored, Message
from ....memory import MemoryStore, MemoryChunk
from ....memory.permanent import (
    PermanentMessage,
    PermanentMessageScored,
    RecordNotFoundException,
    RecordNotSavedException,
)
from logging import Logger
from time import perf_counter
from pgvector.psycopg import register_vector_async
from psycopg_pool import AsyncConnectionPool
from psycopg import AsyncConnection, AsyncCursor
from psycopg.rows import dict_row
from psycopg.types import TypeInfo
from typing import TypeVar

T = TypeVar("T")


class BasePgsqlMemory(MemoryStore[T]):
    _database: AsyncConnection
    _logger: Logger

    def __init__(self, database: AsyncConnectionPool, logger: Logger):
        self._database = database
        self._logger = logger

    async def open(self) -> None:
        await self._database.open()

    async def search(self, query: str) -> list[T] | None:
        raise NotImplementedError()

    async def _execute(
        self, cursor: AsyncCursor, query: str, parameters: tuple | None
    ) -> None:
        self._logger.debug("Executing query: %s with %s", query, parameters)
        start = perf_counter()
        await cursor.execute(query, parameters)
        self._logger.debug(
            "Query finished in %.3f seconds", perf_counter() - start
        )

    async def _fetch_all(
        self, entity: type[T], query: str, parameters: tuple
    ) -> list[T]:
        async with self._database.connection() as connection:
            async with connection.cursor() as cursor:
                await self._execute(cursor, query, parameters)
                results = await cursor.fetchall()
                await cursor.close()
                return (
                    [entity(**dict(result)) for result in results]
                    if results is not None
                    else []
                )

    async def _fetch_one(
        self, entity: type[T], query: str, parameters: tuple
    ) -> T:
        result = await self._try_fetch_one(entity, query, parameters)
        if result is None:
            raise RecordNotFoundException()
        return result

    async def _fetch_field(
        self, field: str, query: str, parameters: tuple | None = None
    ) -> str | None:
        async with self._database.connection() as connection:
            async with connection.cursor() as cursor:
                await self._execute(cursor, query, parameters)
                result = await cursor.fetchone()
                await cursor.close()
                row = dict(result) if result is not None else None
                return row[field] if row else None

    async def _has_one(self, query: str, parameters: tuple) -> bool:
        async with self._database.connection() as connection:
            async with connection.cursor() as cursor:
                await self._execute(cursor, query, parameters)
                result = await cursor.fetchone()
                await cursor.close()
                return result is not None

    async def _try_fetch_one(
        self, entity: type[T], query: str, parameters: tuple
    ) -> T | None:
        async with self._database.connection() as connection:
            async with connection.cursor() as cursor:
                await self._execute(cursor, query, parameters)
                result = await cursor.fetchone()
                await cursor.close()
                return entity(**dict(result)) if result is not None else None

    async def _update_and_fetch_one(
        self, entity: type[T], query: str, parameters: tuple
    ) -> T:
        row = await self._update_and_fetch_row(query, parameters)
        return entity(**row)

    async def _update_and_fetch_field(
        self, field: str, query: str, parameters: tuple
    ) -> str:
        row = await self._update_and_fetch_row(query, parameters)
        return row[field]

    async def _update_and_fetch_row(
        self, query: str, parameters: tuple
    ) -> dict:
        async with self._database.connection() as connection:
            async with connection.cursor() as cursor:
                await self._execute(cursor, query, parameters)
                result = await cursor.fetchone()
                await cursor.close()
                if result is None:
                    raise RecordNotSavedException()
                return dict(result)

    async def _update(self, query: str, parameters: tuple) -> None:
        async with self._database.connection() as connection:
            async with connection.cursor() as cursor:
                await self._execute(cursor, query, parameters)
                await cursor.close()


class PgsqlMemory(BasePgsqlMemory[MemoryChunk[T]]):
    _composite_types: list[str] | None

    @classmethod
    async def create_instance_from_pool(
        cls,
        pool: AsyncConnectionPool,
        *,
        logger: Logger,
        **kwargs,
    ):
        memory = cls(dsn=None, pool=pool, logger=logger, **kwargs)
        return memory

    def __init__(
        self,
        dsn: str | None,
        *args,
        pool: AsyncConnectionPool | None = None,
        composite_types: list[str] | None = None,
        pool_minimum: int | None = None,
        pool_maximum: int | None = None,
        logger: Logger,
        **kwargs,
    ):
        assert pool or (
            dsn
            and pool_minimum
            and pool_minimum
            and pool_minimum > 0
            and pool_maximum > pool_minimum
        )

        if pool:
            super().__init__(database=pool, logger=logger, **kwargs)
        else:
            self._composite_types = composite_types

            if "//" not in dsn:
                dsn = f"postgresql://{dsn}"

            database = AsyncConnectionPool(
                min_size=pool_minimum,
                max_size=pool_maximum,
                conninfo=dsn,
                configure=self._configure_connection,
                open=False,
            )
            super().__init__(database=database, logger=logger, **kwargs)

    async def _configure_connection(self, connection: AsyncConnection):
        connection.row_factory = dict_row
        await connection.set_autocommit(True)
        if self._composite_types:
            for composite_type_name in self._composite_types:
                composite_type = await TypeInfo.fetch(
                    connection, composite_type_name
                )
                if composite_type:
                    composite_type.register(connection)
        await register_vector_async(connection)

    @staticmethod
    def _to_engine_messages(
        messages: list[PermanentMessage] | list[PermanentMessageScored],
        *args,
        limit: int | None,
        reverse: bool = False,
        scored: bool = False,
    ) -> list[EngineMessage] | list[EngineMessageScored]:
        engine_messages = [
            (
                EngineMessageScored(
                    agent_id=m.agent_id,
                    model_id=m.model_id,
                    message=Message(role=m.author, content=m.data),
                    score=m.score,
                )
                if scored
                else EngineMessage(
                    agent_id=m.agent_id,
                    model_id=m.model_id,
                    message=Message(role=m.author, content=m.data),
                )
            )
            for m in messages
        ]
        if reverse:
            engine_messages.reverse()
        if limit and len(engine_messages) > limit:
            engine_messages = engine_messages[:limit]
        return engine_messages



from ...compat import override
from ...model.vendor import TextGenerationVendor
from ...model.audio import BaseAudioModel
from ...model.engine import Engine
from diffusers import DiffusionPipeline
from torch import argmax, inference_mode
from transformers import (
    AutoModelForCTC,
    AutoProcessor,
    PreTrainedModel,
)
from typing import Literal


class SpeechRecognitionModel(BaseAudioModel):
    _processor: AutoProcessor

    def _load_model(
        self,
    ) -> PreTrainedModel | TextGenerationVendor | DiffusionPipeline:
        self._processor = AutoProcessor.from_pretrained(
            self._model_id,
            trust_remote_code=self._settings.trust_remote_code,
            # default behavior in transformers v4.48
            use_fast=True,
            subfolder=self._settings.tokenizer_subfolder or "",
        )
        model = AutoModelForCTC.from_pretrained(
            self._model_id,
            trust_remote_code=self._settings.trust_remote_code,
            pad_token_id=self._processor.tokenizer.pad_token_id,
            ctc_loss_reduction="mean",
            device_map=self._device,
            tp_plan=Engine._get_tp_plan(self._settings.parallel),
            ignore_mismatched_sizes=True,
            subfolder=self._settings.subfolder or "",
        )
        return model

    @override
    async def __call__(
        self,
        path: str,
        sampling_rate: int = 16_000,
        tensor_format: Literal["pt"] = "pt",
    ) -> str:
        audio = self._resample(path, sampling_rate)
        inputs = self._processor(
            audio,
            sampling_rate=sampling_rate,
            return_tensors=tensor_format,
        ).to(self._device)
        with inference_mode():
            # shape (batch, time_steps, vocab_size)
            logits = self._model(inputs.input_values).logits
        predicted_ids = argmax(logits, dim=-1)
        transcription = self._processor.batch_decode(predicted_ids)[0]
        return transcription



from importlib.metadata import metadata, version as metadata_version
from packaging.version import parse, Version
from urllib.parse import urlparse
from urllib.parse import ParseResult


def _config() -> dict:
    config = metadata("avalan")
    package_version = metadata_version("avalan")
    return {
        "name": config["Name"],
        "version": package_version,
        "license": config["License"],
        "url": "https://avalan.ai",
    }


config = _config()


def license() -> str:
    assert "license" in config
    return config["license"]


def name() -> str:
    assert "name" in config
    return config["name"]


def version() -> Version:
    assert "version" in config
    return parse(config["version"])


def site() -> ParseResult:
    assert "url" in config
    return urlparse(config["url"])



from abc import ABC
from asyncio import to_thread as _to_thread
from inspect import iscoroutinefunction
from logging import Logger
from typing import Any, Callable


async def to_thread(func: Callable[..., Any], **kwargs: Any) -> Any:
    """Run *func* in a thread."""
    return await _to_thread(func, **kwargs)


class S3VectorsMemory(ABC):
    """Common logic for S3Vectors based memories."""

    _bucket: str
    _collection: str
    _client: Any
    _logger: Logger

    def __init__(
        self,
        *,
        bucket: str,
        collection: str,
        client: Any,
        logger: Logger,
        **kwargs: Any,
    ) -> None:
        self._bucket = bucket
        self._collection = collection
        self._client = client
        self._logger = logger

    async def _call_client(
        self, func: Callable[..., Any], **kwargs: Any
    ) -> Any:
        if iscoroutinefunction(func):
            return await func(**kwargs)
        return await to_thread(func, **kwargs)

    async def _put_object(self, **kwargs: Any) -> Any:
        return await self._call_client(self._client.put_object, **kwargs)

    async def _get_object(self, **kwargs: Any) -> Any:
        return await self._call_client(self._client.get_object, **kwargs)

    async def _put_vector(self, **kwargs: Any) -> Any:
        return await self._call_client(self._client.put_vector, **kwargs)

    async def _query_vector(self, **kwargs: Any) -> Any:
        return await self._call_client(self._client.query_vector, **kwargs)



from ....vendor import TextGenerationVendor, TextGenerationVendorStream
from . import TextGenerationVendorModel
from .....compat import override
from .....entities import (
    GenerationSettings,
    Message,
    Token,
    TokenDetail,
)
from .....tool.manager import ToolManager
from anthropic import AsyncAnthropic
from anthropic.types import RawContentBlockDeltaEvent, RawMessageStopEvent
from diffusers import DiffusionPipeline
from transformers import PreTrainedModel
from typing import AsyncIterator


class AnthropicStream(TextGenerationVendorStream):
    _stream: AsyncIterator

    def __init__(self, stream: AsyncIterator):
        super().__init__(stream.__aiter__())

    async def __anext__(self) -> Token | TokenDetail | str:
        # We may handle multiple iterations before yielding because
        # Anthropic triggers multiple events besides the deltas
        while True:
            event = await self._generator.__anext__()
            if isinstance(event, RawContentBlockDeltaEvent):
                delta = event.delta
                value = (
                    delta.text
                    if hasattr(delta, "text")
                    else (
                        delta.partial_json
                        if hasattr(delta, "partial_json")
                        else (
                            delta.thinking
                            if hasattr(delta, "thinking")
                            else None
                        )
                    )
                )
                if value is not None:
                    return value
            elif isinstance(event, RawMessageStopEvent):
                raise StopAsyncIteration


class AnthropicClient(TextGenerationVendor):
    _client: AsyncAnthropic

    def __init__(self, api_key: str, base_url: str | None = None):
        self._client = AsyncAnthropic(api_key=api_key, base_url=base_url)

    @override
    async def __call__(
        self,
        model_id: str,
        messages: list[Message],
        settings: GenerationSettings | None = None,
        *,
        tool: ToolManager | None = None,
        use_async_generator: bool = True,
    ) -> AsyncIterator[Token | TokenDetail | str]:
        settings = settings or GenerationSettings()
        system_prompt = self._system_prompt(messages)
        template_messages = self._template_messages(messages, ["system"])
        stream = await self._client.messages.create(
            model=model_id,
            system=system_prompt,
            messages=template_messages,
            max_tokens=settings.max_new_tokens,
            temperature=settings.temperature,
            stream=use_async_generator,
        )
        return AnthropicStream(stream=stream)


class AnthropicModel(TextGenerationVendorModel):
    def _load_model(
        self,
    ) -> TextGenerationVendor | PreTrainedModel | DiffusionPipeline:
        assert self._settings.access_token
        return AnthropicClient(
            api_key=self._settings.access_token,
            base_url=self._settings.base_url,
        )



from ...compat import override
from ...entities import Input
from ...model.vendor import TextGenerationVendor
from ...model.nlp import BaseNLPModel
from ...model.engine import Engine
from diffusers import DiffusionPipeline
from contextlib import nullcontext
from numpy import ndarray
from torch import inference_mode
from transformers import PreTrainedModel
from transformers.tokenization_utils_base import BatchEncoding
from typing import Literal


class SentenceTransformerModel(BaseNLPModel):
    @property
    def supports_sample_generation(self) -> bool:
        return False

    @property
    def supports_token_streaming(self) -> bool:
        return False

    @property
    def uses_tokenizer(self) -> bool:
        return True

    def token_count(self, input: str) -> int:
        token_ids = self.tokenizer.encode(input, add_special_tokens=False)
        return len(token_ids)

    def _load_model(
        self,
    ) -> PreTrainedModel | TextGenerationVendor | DiffusionPipeline:
        from sentence_transformers import SentenceTransformer

        model = SentenceTransformer(
            self._model_id,
            cache_folder=self._settings.cache_dir,
            device=self._device,
            trust_remote_code=self._settings.trust_remote_code,
            local_files_only=self._settings.local_files_only,
            token=self._settings.access_token,
            model_kwargs={
                "attn_implementation": self._settings.attention,
                "torch_dtype": Engine.weight(self._settings.weight_type),
                "low_cpu_mem_usage": (
                    True if self._device else self._settings.low_cpu_mem_usage
                ),
                "device_map": self._device,
                "tp_plan": Engine._get_tp_plan(self._settings.parallel),
            },
            backend="torch",
            similarity_fn_name=None,
            truncate_dim=None,
        )
        return model

    def _tokenize_input(
        self,
        input: Input,
        system_prompt: str | None,
        context: str | None,
        tensor_format: Literal["pt"] = "pt",
        chat_template_settings: dict[str, object] | None = None,
    ) -> BatchEncoding:
        raise NotImplementedError()

    @override
    async def __call__(
        self, input: Input, *args, enable_gradient_calculation: bool = False
    ) -> ndarray:
        assert self._model, (
            f"Model {self._model} can't be executed, it "
            + "needs to be loaded first"
        )
        assert isinstance(input, str) or isinstance(input, list)
        with (
            inference_mode()
            if not enable_gradient_calculation
            else nullcontext()
        ):
            embeddings = self._model.encode(input, convert_to_numpy=True)
        return embeddings



from ....compat import override
from ....entities import EngineSettings, Input
from ....model.vendor import TextGenerationVendor
from ....model.engine import Engine
from ....model.vision import BaseVisionModel
from dataclasses import replace
from diffusers import DiffusionPipeline
from diffusers.pipelines.ltx.pipeline_ltx_condition import LTXVideoCondition
from diffusers.utils import export_to_video, load_image, load_video
from logging import Logger, getLogger
from torch import Generator, inference_mode
from transformers import PreTrainedModel


class TextToVideoModel(BaseVisionModel):
    _upsampler_pipe: DiffusionPipeline

    def __init__(
        self,
        model_id: str,
        settings: EngineSettings | None = None,
        logger: Logger = getLogger(__name__),
    ):
        settings = settings or EngineSettings()
        assert settings.upsampler_model_id
        settings = replace(settings, enable_eval=False)
        super().__init__(model_id, settings, logger)

    def _load_model(
        self,
    ) -> PreTrainedModel | TextGenerationVendor | DiffusionPipeline:
        dtype = Engine.weight(self._settings.weight_type)
        base_pipe = DiffusionPipeline.from_pretrained(
            self._model_id,
            torch_dtype=dtype,
        ).to(self._device)
        self._upsampler_pipe = DiffusionPipeline.from_pretrained(
            self._settings.upsampler_model_id,
            vae=base_pipe.vae,
            torch_dtype=dtype,
        ).to(self._device)
        base_pipe.vae.enable_tiling()
        return base_pipe

    @override
    async def __call__(
        self,
        input: Input,
        negative_prompt: str,
        reference_path: str,
        path: str,
        *,
        decode_timestep: float = 0.05,
        denoise_strength: float = 0.4,
        downscale: float = 2 / 3,
        fps: int = 24,
        frames: int = 96,
        height: int = 480,
        inference_steps: int = 10,
        noise_scale: float = 0.025,
        width: int = 832,
        steps: int = 30,
    ) -> str:
        image = load_image(reference_path)
        video = load_video(export_to_video([image]))
        condition = LTXVideoCondition(video=video, frame_index=0)

        down_h = int(height * downscale)
        down_w = int(width * downscale)
        down_h, down_w = (
            TextToVideoModel._round_to_nearest_resolution_acceptable_by_vae(
                down_h,
                down_w,
                ratio=self._model.vae_spatial_compression_ratio,
            )
        )

        with inference_mode():
            latents = self._model(
                conditions=[condition],
                prompt=input if isinstance(input, str) else str(input),
                negative_prompt=negative_prompt,
                width=down_w,
                height=down_h,
                num_frames=frames,
                num_inference_steps=steps,
                generator=Generator().manual_seed(0),
                output_type="latent",
            ).frames

            upscaled_h, upscaled_w = down_h * 2, down_w * 2
            upscaled_latents = self._upsampler_pipe(
                latents=latents, output_type="latent"
            ).frames

            video = self._model(
                conditions=[condition],
                prompt=input if isinstance(input, str) else str(input),
                negative_prompt=negative_prompt,
                width=upscaled_w,
                height=upscaled_h,
                num_frames=frames,
                denoise_strength=denoise_strength,
                num_inference_steps=inference_steps,
                latents=upscaled_latents,
                decode_timestep=decode_timestep,
                image_cond_noise_scale=noise_scale,
                generator=Generator().manual_seed(0),
                output_type="pil",
            ).frames[0]

        video = [frame.resize((width, height)) for frame in video]
        export_to_video(video, path, fps=fps)

        return path

    @staticmethod
    def _round_to_nearest_resolution_acceptable_by_vae(
        height: int,
        width: int,
        ratio: int,
    ) -> tuple[int, int]:
        return (
            height - (height % ratio),
            width - (width % ratio),
        )



from __future__ import annotations

from typing import Callable, TypeVar

try:
    from typing import override as _override
except ImportError:  # Python < 3.12
    T = TypeVar("T", bound=Callable[..., object])

    def _override(func: T) -> T:  # type: ignore
        return func


override = _override

__all__ = ["override"]



from . import Tool


class SearchEngineTool(Tool):
    """
    Can search Internet search engines for real time information.

    Args:
        query: Term to search for. Example: "weather today in Buenos Aires"
        engine: Search engine to use. Example: "google"

    Returns:
        Result of searching Google for the given term.
    """

    def __init__(self) -> None:
        self.__name__ = "search"

    async def __call__(self, query: str, engine: str) -> str:
        return (
            "The weather is nice and warm, with 23 degrees celsius, clear"
            " skies, and winds under 11 kmh."
        )



from dataclasses import dataclass
from enum import StrEnum
from typing import Any


class EventType(StrEnum):
    CALL_PREPARE_BEFORE = "call_prepare_before"
    CALL_PREPARE_AFTER = "call_prepare_after"
    END = "end"
    ENGINE_RUN_BEFORE = "engine_run_before"
    ENGINE_RUN_AFTER = "engine_run_after"
    INPUT_TOKEN_COUNT_BEFORE = "input_token_count_before"
    INPUT_TOKEN_COUNT_AFTER = "input_token_count_after"
    MEMORY_APPEND_BEFORE = "memory_append_before"
    MEMORY_APPEND_AFTER = "memory_append_after"
    MEMORY_PERMANENT_MESSAGE_ADD = "memory_permanent_message_add"
    MEMORY_PERMANENT_MESSAGE_ADDED = "memory_permanent_message_added"
    MEMORY_PERMANENT_MESSAGE_SESSION_CONTINUE = (
        "memory_permanent_message_session_continue"
    )
    MEMORY_PERMANENT_MESSAGE_SESSION_CONTINUED = (
        "memory_permanent_message_session_continued"
    )
    MEMORY_PERMANENT_MESSAGE_SESSION_START = (
        "memory_permanent_message_session_start"
    )
    MEMORY_PERMANENT_MESSAGE_SESSION_STARTED = (
        "memory_permanent_message_session_started"
    )
    MODEL_EXECUTE_BEFORE = "model_execute_before"
    MODEL_EXECUTE_AFTER = "model_execute_after"
    MODEL_MANAGER_CALL_BEFORE = "model_manager_call_before"
    MODEL_MANAGER_CALL_AFTER = "model_manager_call_after"
    FLOW_MANAGER_CALL_BEFORE = "flow_manager_call_before"
    FLOW_MANAGER_CALL_AFTER = "flow_manager_call_after"
    START = "start"
    STREAM_END = "stream_end"
    TOKEN_GENERATED = "token_generated"
    TOOL_DETECT = "tool_detect"
    TOOL_EXECUTE = "tool_execute"
    TOOL_MODEL_RUN = "tool_model_run"
    TOOL_MODEL_RESPONSE = "tool_model_response"
    TOOL_PROCESS = "tool_process"
    TOOL_RESULT = "tool_result"


TOOL_TYPES = {et for et in EventType if et.value.startswith("tool_")}


@dataclass(frozen=True, kw_only=True)
class Event:
    type: EventType
    payload: dict[str, Any] | None = None
    started: float | None = None
    finished: float | None = None
    elapsed: float | None = None


class EventStats:
    triggers: dict[EventType, int] = {}
    total_triggers: int = 0



from dataclasses import dataclass, field
from datetime import datetime
from enum import StrEnum
from collections.abc import Callable
from numpy import ndarray
from torch import dtype, Tensor
from typing import Literal, TypedDict
from uuid import UUID

AttentionImplementation = Literal[
    "eager", "flash_attention_2", "flex_attention", "sdpa"
]

ProbabilityDistribution = Literal[
    "entmax", "gumbel_softmax", "log_softmax", "sparsemax", "softmax"
]

SimilarityFunction = Literal["cosine", "dot", "euclidean", "manhattan"]

ImageTextGenerationLoaderClass = Literal["gemma3", "qwen2"]

TextGenerationLoaderClass = Literal["auto", "gemma3", "mistral3"]


class Backend(StrEnum):
    TRANSFORMERS = "transformers"
    MLXLM = "mlxlm"
    VLLM = "vllm"


ToolValue = bool | float | int | str | None

Vendor = Literal[
    "anthropic",
    "anyscale",
    "deepinfra",
    "deepseek",
    "google",
    "groq",
    "huggingface",
    "hyperbolic",
    "local",
    "openai",
    "openrouter",
    "ollama",
    "litellm",
    "together",
]

WeightType = Literal[
    "auto",
    "bool",
    "bf16",
    "f16",
    "f32",
    "f64",
    "fp16",
    "fp32",
    "i8",
    "i16",
    "i32",
    "i64",
    "ui8",
]


class DistanceType(StrEnum):
    COSINE = "cosine"
    DOT = "dot"
    L1 = "l1"
    L2 = "l2"
    PEARSON = "pearson"


class MessageRole(StrEnum):
    ASSISTANT = "assistant"
    SYSTEM = "system"
    TOOL = "tool"
    USER = "user"


class Modality(StrEnum):
    AUDIO_CLASSIFICATION = "audio_classification"
    AUDIO_SPEECH_RECOGNITION = "audio_speech_recognition"
    AUDIO_TEXT_TO_SPEECH = "audio_text_to_speech"
    AUDIO_GENERATION = "audio_generation"
    EMBEDDING = "embedding"
    TEXT_GENERATION = "text_generation"
    TEXT_QUESTION_ANSWERING = "text_question_answering"
    TEXT_SEQUENCE_CLASSIFICATION = "text_sequence_classification"
    TEXT_SEQUENCE_TO_SEQUENCE = "text_sequence_to_sequence"
    TEXT_TRANSLATION = "text_translation"
    TEXT_TOKEN_CLASSIFICATION = "text_token_classification"
    VISION_OBJECT_DETECTION = "vision_object_detection"
    VISION_IMAGE_CLASSIFICATION = "vision_image_classification"
    VISION_IMAGE_TO_TEXT = "vision_image_to_text"
    VISION_TEXT_TO_IMAGE = "vision_text_to_image"
    VISION_TEXT_TO_ANIMATION = "vision_text_to_animation"
    VISION_TEXT_TO_VIDEO = "vision_text_to_video"
    VISION_IMAGE_TEXT_TO_TEXT = "vision_image_text_to_text"
    VISION_ENCODER_DECODER = "vision_encoder_decoder"
    VISION_SEMANTIC_SEGMENTATION = "vision_semantic_segmentation"


class ParallelStrategy(StrEnum):
    AUTO = "auto"
    COLWISE = "colwise"
    ROWWISE = "rowwise"
    COLWISE_REP = "colwise_rep"
    ROWWISE_REP = "rowwise_rep"
    LOCAL_COLWISE = "local_colwise"
    LOCAL_ROWWISE = "local_rowwise"
    LOCAL = "local"
    GATHER = "gather"
    LOCAL_PACKED_ROWWISE = "local_packed_rowwise"
    SEQUENCE_PARALLEL = "sequence_parallel"
    REPLICATE = "replicate"


class ToolFormat(StrEnum):
    JSON = "json"
    REACT = "react"
    BRACKET = "bracket"
    OPENAI = "openai"


class VisionColorModel(StrEnum):
    ONE = "1"
    L = "L"
    LA = "LA"
    P = "P"
    PA = "PA"
    RGB = "RGB"
    RGBA = "RGBA"
    RGBX = "RGBX"
    CMYK = "CMYK"
    YCBCR = "YCbCr"
    LAB = "LAB"
    HSV = "HSV"
    INTENSITY = "I"
    FLOAT = "F"


class VisionImageFormat(StrEnum):
    BMP = "BMP"
    DDS = "DDS"
    EPS = "EPS"
    GIF = "GIF"
    ICNS = "ICNS"
    ICO = "ICO"
    IM = "IM"
    JPEG = "JPEG"
    JPEG2000 = "JPEG2000"
    MSP = "MSP"
    PCX = "PCX"
    PNG = "PNG"
    PPM = "PPM"
    SGI = "SGI"
    SPI = "SPI"
    TGA = "TGA"
    TIFF = "TIFF"
    WEBP = "WEBP"
    XBM = "XBM"


class TimestepSpacing(StrEnum):
    LINSPACE = "linspace"
    LEADING = "leading"
    TRAILING = "trailing"


class BetaSchedule(StrEnum):
    LINEAR = "linear"
    SCALED_LINEAR = "scaled_linear"
    SQUAREDCOS_CAP_V2 = "squaredcos_cap_v2"


@dataclass(frozen=True, kw_only=True)
class QuantizationSettings:
    load_in_4bit: bool
    bnb_4bit_quant_type: Literal["nf4"]
    bnb_4bit_use_double_quant: bool
    bnb_4bit_compute_dtype: type


@dataclass(frozen=True, kw_only=True)
class EngineSettings:
    auto_load_model: bool = True
    auto_load_tokenizer: bool = True
    cache_dir: str | None = None
    change_transformers_logging_level = True
    device: str | None = None
    disable_loading_progress_bar: bool = True
    enable_eval: bool = True
    parallel: ParallelStrategy | dict[str, ParallelStrategy] | None = None
    trust_remote_code: bool = False
    tokenizer_name_or_path: str | None = None
    subfolder: str | None = None
    tokenizer_subfolder: str | None = None
    access_token: str | None = None
    base_url: str | None = None
    revision: str | None = None
    quantization: QuantizationSettings | None = None
    weight_type: WeightType = "auto"
    base_model_id: str | None = None
    checkpoint: str | None = None
    refiner_model_id: str | None = None
    upsampler_model_id: str | None = None


@dataclass(kw_only=True, frozen=True)
class EngineUri:
    host: str | None
    port: int | None
    user: str | None
    password: str | None
    vendor: Vendor | None
    model_id: str | None
    params: dict[str, str | int | float | bool]

    @property
    def is_local(self) -> bool:
        return not self.vendor or self.vendor == "local"


@dataclass(frozen=True, kw_only=True)
class ReasoningSettings:
    max_new_tokens: int | None = None
    enabled: bool = True
    stop_on_max_new_tokens: bool = False


@dataclass(frozen=True, kw_only=True)
class ChatSettings:
    add_generation_prompt: bool = True
    tokenize: bool = True
    add_special_tokens: bool = True
    return_dict: bool = True
    enable_thinking: bool = True


@dataclass(frozen=True, kw_only=True)
class GenerationSettings:
    # Generation length ------------------------------------------------------
    # The minimum numbers of tokens to generate, ignoring the number of tokens
    # in the prompt.
    min_new_tokens: int | None = None
    # The maximum numbers of tokens to generate, ignoring the number of tokens
    # in the prompt
    max_new_tokens: int | None = None
    # The minimum length of the sequence to be generated. Corresponds to the
    # length of the input prompt + min_new_tokens. Its effect is overridden
    # by min_new_tokens, if also set.
    min_length: int | None = 0
    # The maximum length the generated tokens can have. Corresponds to the
    # length of the input prompt + max_new_tokens. Its effect is overridden
    # by max_new_tokens, if also set.
    max_length: int | None = 20
    # Controls the stopping condition for beam-based methods, like
    # beam-search. It accepts the following values: True, where the
    # generation stops as soon as there are num_beams complete candidates;
    # False, where an heuristic is applied and the generation stops when is
    # it very unlikely to find better candidates; "never", where the beam
    # search procedure only stops when there cannot be better candidates
    # (canonical beam search algorithm)
    early_stopping: str | bool | None = False
    # The maximum amount of time you allow the computation to run for in
    # seconds. generation will still finish the current pass after allocated
    # time has been passed.
    max_time: float | None = None
    # A string or a list of strings that should terminate generation if the
    # model outputs them.
    stop_strings: str | list[str] | None = None

    # Generation strategy ----------------------------------------------------
    # Whether or not to use sampling. Use greedy decoding otherwise
    do_sample: bool = False
    # Number of beams for beam search. 1 means no beam search.
    num_beams: int | None = 1
    # Number of groups to divide num_beams into in order to ensure diversity
    # among different groups of beams
    num_beam_groups: int | None = 1
    # The values balance the model confidence and the degeneration penalty in
    # contrastive search decoding
    penalty_alpha: float | None = None

    # Cache ------------------------------------------------------------------
    # Whether or not the model should use the past last key/values attentions
    # (if applicable to the model) to speed up decoding
    use_cache: bool = True

    # Generation output variables --------------------------------------------
    # The number of independently computed returned sequences for each element
    # in the batch.
    num_return_sequences: int | None = 1
    # Whether or not to return the attentions tensors of all attention layers.
    # See attentions under returned tensors for more details.
    output_attentions: bool | None = False
    # Whether or not to return the hidden states of all layers. See
    # hidden_states under returned tensors for more details.
    output_hidden_states: bool | None = False
    # Whether or not to return the prediction scores. See scores under
    # returned tensors for more details.
    output_scores: bool | None = False
    # Whether or not to return the unprocessed prediction logit scores. See
    # logits under returned tensors for more details.
    output_logits: bool | None = None
    # Whether or not to return a ModelOutput, as opposed to returning
    # exclusively the generated sequence. This flag must be set to True to
    # return the generation cache (when use_cache is True) or optional
    # outputs (see flags starting with output_)
    return_dict_in_generate: bool | None = False

    # Output logits manipulation ---------------------------------------------
    # The value used to module the next token probabilities
    temperature: float | None = 1.0
    # The number of highest probability vocabulary tokens to keep for
    # top-k-filtering
    top_k: int | None = 50
    # If set to float < 1, only the smallest set of most probable tokens with
    # probabilities that add up to top_p or higher are kept for generation
    top_p: float | None = 1.0
    # Minimum token probability, which will be scaled by the probability of
    # the most likely token. It must be a value between 0 and 1. Typical
    # values are in the 0.01-0.2 range, comparably selective as setting
    # top_p in the 0.99-0.8 range (use the opposite of normal top_p values)
    min_p: float | None = None
    # The parameter for repetition penalty. 1.0 means no penalty
    repetition_penalty: float | None = 1.0
    # This value is subtracted from a beam’s score if it generates a token
    # same as any beam from other group at a particular time. Note that
    # diversity_penalty is only effective if group beam search is enabled
    diversity_penalty: float | None = 0.0
    # The id of the token to force as the first generated token after the
    # decoder_start_token_id. Useful for multilingual models like mBART
    # where the first generated token needs to be the target language token.
    forced_bos_token_id: int | None = None
    # The id of the token to force as the last generated token when 3
    # max_length is reached. Optionally, use a list to set multiple
    # end-of-sequence tokens
    forced_eos_token_id: int | list[int] | None = None

    # Special token usage in generation --------------------------------------
    # The id of the padding token.
    pad_token_id: int | None = None
    # The id of the beginning-of-sequence token
    bos_token_id: int | None = None
    # The id of the end-of-sequence token. Optionally, use a list to set
    # multiple end-of-sequence tokens
    eos_token_id: int | list[int] | None = None

    # Assistant generation ---------------------------------------------------
    # The number of tokens to be output as candidate tokens.
    prompt_lookup_num_tokens: int | None = None

    # Inference settings -----------------------------------------------------
    # Gradient calculation (set to true when torch.backwards() is called)
    enable_gradient_calculation: bool = False
    # Use async generator (token streaming)
    use_async_generator: bool = True
    # Use the attention mask discovered during input tokenization
    use_inputs_attention_mask: bool = True
    # Parameters passed to tokenizer.apply_chat_template
    chat_settings: ChatSettings = field(default_factory=ChatSettings)
    reasoning: ReasoningSettings = field(default_factory=ReasoningSettings)


@dataclass(frozen=True, kw_only=True)
class GenericProxyConfig:
    scheme: str
    host: str
    port: int
    username: str | None = None
    password: str | None = None

    def to_dict(self) -> dict[str, str]:
        credentials = (
            f"{self.username}:{self.password}@"
            if self.username and self.password
            else ""
        )
        url = f"{self.scheme}://{credentials}{self.host}:{self.port}"
        return {"http": url, "https": url}


@dataclass(frozen=True, kw_only=True)
class HubCacheFile:
    name: str
    path: str
    size_on_disk: int
    last_accessed: datetime
    last_modified: datetime


@dataclass(frozen=True, kw_only=True)
class HubCache:
    model_id: str
    path: str
    size_on_disk: int
    revisions: list[str]
    files: dict[str, list[HubCacheFile]]
    total_files: int
    total_revisions: int


@dataclass(frozen=True, kw_only=True)
class HubCacheDeletion:
    model_id: str
    revisions: list[str]
    deletable_size_on_disk: float
    deletable_blobs: list[str]
    deletable_refs: list[str]
    deletable_repos: list[str]
    deletable_snapshots: list[str]


@dataclass(frozen=True, kw_only=True)
class ImageEntity:
    label: str
    score: float | None = None
    box: list[float] | None = None


@dataclass(frozen=True, kw_only=True)
class MessageContentText:
    type: Literal["text"]
    text: str


@dataclass(frozen=True, kw_only=True)
class MessageContentImage:
    type: Literal["image_url"]
    image_url: dict[str, str]


MessageContent = MessageContentText | MessageContentImage


@dataclass(frozen=True, kw_only=True)
class Message:
    role: MessageRole
    content: str | MessageContent | list[MessageContent]
    name: str | None = None
    arguments: dict | None = None


Input = str | list[str] | Message | list[Message]


@dataclass(frozen=True, kw_only=True)
class ReasoningOrchestratorResponse:
    """Result returned by :class:`ReasoningOrchestrator`."""

    answer: str
    reasoning: str | None = None


@dataclass(frozen=True, kw_only=True)
class EngineMessage:
    agent_id: UUID
    model_id: str
    message: Message

    @property
    def is_from_agent(self) -> bool:
        return self.message.role == MessageRole.ASSISTANT


@dataclass(frozen=True, kw_only=True)
class EngineMessageScored(EngineMessage):
    score: float


@dataclass(frozen=True, kw_only=True)
class Model:
    id: str
    parameters: int | None
    parameter_types: list[str] | None
    inference: str | None
    library_name: str | None
    license: str | None
    pipeline_tag: str | None
    tags: list[str]
    architectures: list[str] | None
    model_type: str | None
    auto_model: str | None
    processor: str | None
    gated: Literal["auto", "manual", False] | None
    private: bool
    disabled: bool | None
    last_downloads: int
    downloads: int
    likes: int
    ranking: int | None
    author: str
    created_at: datetime
    updated_at: datetime


@dataclass(frozen=True, kw_only=True)
class ModelConfig:
    # Model architectures that can be used with the model pretrained weights
    architectures: list[str] | None
    # A dict that maps model specific attribute names to the standardized
    # naming of attributes
    attribute_map: dict[str, str]
    # The id of the beginning-of-stream token
    bos_token_id: str | None
    bos_token: str | None
    # If an encoder-decoder model starts decoding with a different token
    # than bos, the id of that token
    decoder_start_token_id: int | None
    # The id of the end-of-stream token
    eos_token_id: int | None
    eos_token: str | None
    # Name of the task used to fine-tune the model. This can be used when
    # converting from an original (TensorFlow or PyTorch) checkpoint
    finetuning_task: str | None
    # The hidden size of the model
    hidden_size: int | None
    # The hidden sizes of the model (ResNet)
    hidden_sizes: list[int] | None
    # A list of keys to ignore by default when looking at dictionary outputs
    # of the model during inference
    keys_to_ignore_at_inference: list[str]
    # The type of loss that the model should use
    loss_type: str | None
    # Maximum input sequence length
    max_position_embeddings: int | None
    # An identifier for the model type
    model_type: str | None
    # The number of attention heads used in the multi-head attention layers
    # of the model
    num_attention_heads: int | None
    # The number of blocks in the model
    num_hidden_layers: int | None
    # Number of labels to use in the last layer added to the model, typically
    # for a classification task
    num_labels: int | None
    # Whether or not the model should returns all attentions
    output_attentions: bool
    # Whether or not the model should return all hidden-states
    output_hidden_states: bool
    # The id of the padding token
    pad_token_id: int | None
    pad_token: str | None
    # A specific prompt that should be added at the beginning of each text
    # before calling the model
    prefix: str | None
    # The id of the separation token
    sep_token_id: int | None
    sep_token: str | None
    state_size: int
    # Additional keyword arguments to store for the current task
    task_specific_params: dict[str, any] | None
    # The dtype of the weight. Since the config object is stored in plain
    # text, this attribute contains just the floating type string without the
    # torch
    torch_dtype: dtype
    # The number of tokens in the vocabulary, which is also the first
    # dimension of the embeddings matrix
    vocab_size: int | None
    # The name of the associated tokenizer class to use (if none is set, will
    # use the tokenizer associated to the model by default)
    tokenizer_class: str | None


@dataclass(frozen=True, kw_only=True)
class OrchestratorSettings:
    agent_id: UUID
    orchestrator_type: str | None
    agent_config: dict
    uri: str
    engine_config: dict
    call_options: dict | None
    template_vars: dict | None
    memory_permanent_message: str | None
    permanent_memory: dict[str, str] | None
    memory_recent: bool
    sentence_model_id: str
    sentence_model_engine_config: dict | None
    sentence_model_max_tokens: int
    sentence_model_overlap_size: int
    sentence_model_window_size: int
    json_config: dict | None
    tools: list[str]
    log_events: bool


@dataclass(frozen=True, kw_only=True)
class SearchMatch:
    query: str
    match: str
    l2_distance: float


@dataclass(frozen=True, kw_only=True)
class SentenceTransformerModelConfig:
    backend: Literal["torch", "onnx", "openvino"]
    similarity_function: SimilarityFunction | None
    truncate_dimension: int | None
    transformer_model_config: ModelConfig


@dataclass(frozen=True, kw_only=True)
class Similarity:
    cosine_distance: float
    inner_product: float
    l1_distance: float
    l2_distance: float
    pearson: float


@dataclass(frozen=True, kw_only=True)
class OperationAudioParameters:
    path: str
    reference_path: str | None = None
    reference_text: str | None = None
    sampling_rate: int


@dataclass(frozen=True, kw_only=True)
class OperationTextParameters:
    context: str | None = None
    labeled_only: bool | None = None
    language_destination: str | None = None
    language_source: str | None = None
    manual_sampling: bool | None = None
    pick_tokens: int | None = None
    skip_special_tokens: bool | None = None
    stop_on_keywords: list[str] | None = None
    system_prompt: str | None = None


@dataclass(frozen=True, kw_only=True)
class OperationVisionParameters:
    path: str
    reference_path: str | None = None
    negative_prompt: str | None = None
    height: int | None = None
    downscale: float | None = 2 / 3
    frames: int | None = 96
    denoise_strength: float | None = 0.4
    inference_steps: int | None = 10
    decode_timestep: float | None = 0.05
    noise_scale: float | None = 0.025
    frames_per_second: int | None = 24
    skip_special_tokens: bool | None = None
    system_prompt: str | None = None
    threshold: float | None = None
    width: int | None = None
    color_model: VisionColorModel | None = None
    high_noise_frac: float | None = None
    image_format: VisionImageFormat | None = None
    n_steps: int | None = None
    timestep_spacing: TimestepSpacing | None = None
    beta_schedule: BetaSchedule | None = None
    guidance_scale: float | None = None


class OperationParameters(TypedDict, total=False):
    audio: OperationAudioParameters | None = None
    text: OperationTextParameters | None = None
    vision: OperationVisionParameters | None = None


@dataclass(frozen=True, kw_only=True)
class Operation:
    generation_settings: GenerationSettings | None
    input: Input | None
    modality: Modality
    parameters: OperationParameters
    requires_input: bool = False


@dataclass(frozen=True, kw_only=True)
class TokenizerConfig:
    name_or_path: str
    tokens: list[str] | None
    special_tokens: list[str] | None
    # Maximum sequence length the tokenizer supports
    tokenizer_model_max_length: int
    # Wether tokenizer is a Rust-based tokenizer
    fast: bool = False


@dataclass(frozen=True, kw_only=True)
class Token:
    id: Tensor | int
    token: str
    probability: float | None = None


@dataclass(frozen=True, kw_only=True)
class TokenDetail(Token):
    step: int | None = None
    probability_distribution: ProbabilityDistribution | None = None
    tokens: list[Token] | None = None


@dataclass(frozen=True, kw_only=True)
class ReasoningToken(Token):
    """Token produced while the model is reasoning."""

    def __init__(
        self,
        token: str,
        *,
        id: Tensor | int = -1,
        probability: float | None = None,
    ) -> None:
        super().__init__(id=id, token=token, probability=probability)


@dataclass(frozen=True, kw_only=True)
class ToolCallToken(Token):
    """Token produced while the model emits a tool call."""

    def __init__(
        self,
        token: str,
        *,
        id: Tensor | int = -1,
        probability: float | None = None,
    ) -> None:
        super().__init__(id=id, token=token, probability=probability)


@dataclass(frozen=True, kw_only=True)
class ToolCall:
    id: UUID
    name: str
    arguments: dict[str, ToolValue] | None = None


@dataclass(frozen=True, kw_only=True)
class ToolCallContext:
    agent_id: UUID | None = None
    input: Input | None = None
    participant_id: UUID | None = None
    session_id: UUID | None = None
    calls: list[ToolCall] | None = None


@dataclass(frozen=True, kw_only=True)
class ToolCallResult(ToolCall):
    id: UUID
    call: ToolCall
    result: ToolValue | None = None


@dataclass(frozen=True, kw_only=True)
class ToolFilter:
    func: Callable[
        [ToolCall, ToolCallContext], tuple[ToolCall, ToolCallContext] | None
    ]
    namespace: str | None = None


@dataclass(frozen=True, kw_only=True)
class ToolTransformer:
    func: Callable[
        [ToolCall, ToolCallContext, ToolValue | None], ToolValue | None
    ]
    namespace: str | None = None


@dataclass(frozen=True, kw_only=True)
class ToolManagerSettings:
    eos_token: str | None = None
    tool_format: ToolFormat | None = None
    avoid_repetition: bool = False
    maximum_depth: int | None = None
    filters: (
        list[
            Callable[
                [ToolCall, ToolCallContext],
                tuple[ToolCall, ToolCallContext] | None,
            ]
            | ToolFilter
        ]
        | None
    ) = None
    transformers: (
        list[
            Callable[
                [ToolCall, ToolCallContext, ToolValue | None], ToolValue | None
            ]
            | ToolTransformer
        ]
        | None
    ) = None


@dataclass(frozen=True, kw_only=True)
class TextPartition:
    data: str
    total_tokens: int
    embeddings: ndarray


@dataclass(frozen=True, kw_only=True)
class TransformerEngineSettings(EngineSettings):
    attention: AttentionImplementation | None = None
    backend: Backend = Backend.TRANSFORMERS
    loader_class: TextGenerationLoaderClass | None = "auto"
    local_files_only: bool = False
    low_cpu_mem_usage: bool = False
    output_hidden_states: bool = False
    special_tokens: list[str] | None = None
    state_dict: dict[str, Tensor] = None
    tokens: list[str] | None = None


@dataclass(frozen=True, kw_only=True)
class User:
    name: str
    full_name: str | None = None
    access_token_name: str | None = None


@dataclass(frozen=True, kw_only=True)
class WebshareProxyConfig:
    host: str
    port: int
    username: str
    password: str
    scheme: str = "http"

    def to_generic(self) -> GenericProxyConfig:
        return GenericProxyConfig(
            scheme=self.scheme,
            host=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
        )



from ...vendor import TextGenerationVendorStream
from ....model.response.text import TextGenerationResponse
from .generation import TextGenerationModel
from ....compat import override
from ....entities import (
    GenerationSettings,
    Input,
    TransformerEngineSettings,
)
from ....tool.manager import ToolManager
from asyncio import to_thread
from dataclasses import asdict, replace
from logging import Logger, getLogger
from mlx_lm import generate, load, stream_generate
from mlx_lm.sample_utils import make_sampler
from torch import Tensor
from typing import AsyncGenerator, Callable, Literal


class MlxLmStream(TextGenerationVendorStream):
    """Async wrapper around a synchronous token generator."""

    _SENTINEL = object()

    def __init__(self, generator):
        super().__init__(generator)
        self._iterator = generator

    async def __anext__(self) -> str:
        sentinel = type(self)._SENTINEL
        chunk = await to_thread(next, self._iterator, sentinel)
        if chunk is sentinel:
            raise StopAsyncIteration
        return chunk


class MlxLmModel(TextGenerationModel):
    def __init__(
        self,
        model_id: str,
        settings: TransformerEngineSettings | None = None,
        logger: Logger = getLogger(__name__),
    ) -> None:
        settings = settings or TransformerEngineSettings()
        if settings.auto_load_tokenizer:
            settings = replace(settings, auto_load_tokenizer=False)
        super().__init__(model_id, settings, logger)

    @property
    def supports_sample_generation(self) -> bool:
        return False

    def _load_model(self):
        model, tokenizer = load(self._model_id)
        self._tokenizer = tokenizer
        self._loaded_tokenizer = True
        return model

    async def _stream_generator(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        skip_special_tokens: bool,
    ) -> AsyncGenerator[str, None]:
        sampler, prompt = self._get_sampler_and_prompt(
            inputs, settings, skip_special_tokens
        )
        iterator = stream_generate(
            self._model,
            self._tokenizer,
            prompt,
            sampler=sampler,
            max_tokens=settings.max_new_tokens,
        )
        stream = MlxLmStream(iter(iterator))
        async for chunk in stream:
            yield chunk.text

    def _string_output(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        skip_special_tokens: bool,
    ) -> str:
        sampler, prompt = self._get_sampler_and_prompt(
            inputs, settings, skip_special_tokens
        )
        return generate(
            self._model,
            self._tokenizer,
            prompt,
            sampler=sampler,
            max_tokens=settings.max_new_tokens,
        )

    @override
    async def __call__(
        self,
        input: Input,
        system_prompt: str | None = None,
        settings: GenerationSettings | None = None,
        *,
        skip_special_tokens: bool = False,
        tensor_format: Literal["pt"] = "pt",
        tool: ToolManager | None = None,
    ) -> TextGenerationResponse:
        settings = settings or GenerationSettings()
        inputs = super()._tokenize_input(
            input,
            system_prompt,
            context=None,
            tensor_format=tensor_format,
            tool=tool,
            chat_template_settings=asdict(settings.chat_settings),
        )
        generation_settings = replace(settings, do_sample=False)
        output_fn = (
            self._stream_generator
            if settings.use_async_generator
            else self._string_output
        )

        return TextGenerationResponse(
            output_fn,
            inputs=inputs,
            logger=self._logger,
            generation_settings=generation_settings,
            settings=generation_settings,
            skip_special_tokens=skip_special_tokens,
            use_async_generator=settings.use_async_generator,
        )

    def _get_sampler_and_prompt(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        skip_special_tokens: bool,
    ) -> tuple[Callable, str]:
        sampler_settings = {
            "temp": settings.temperature,
            "top_p": settings.top_p,
            "top_k": settings.top_k,
        }
        sampler_settings = {
            k: v for k, v in sampler_settings.items() if v is not None
        }
        sampler = make_sampler(**sampler_settings)
        prompt = self._tokenizer.decode(
            inputs["input_ids"][0],
            skip_special_tokens=skip_special_tokens,
        )
        return sampler, prompt



from abc import ABC
from ...model.engine import Engine
from PIL import Image


class BaseVisionModel(Engine, ABC):
    @staticmethod
    def _get_image(image_source: str | Image.Image) -> Image.Image:
        return (
            image_source
            if isinstance(image_source, Image.Image)
            else Image.open(image_source).convert("RGB")
        )



from asyncio import AbstractEventLoop, get_running_loop
from botocore.exceptions import ClientError
from boto3 import client
from boto3.session import Session
from concurrent.futures import ThreadPoolExecutor
from typing import Callable


class DeployError(Exception):
    """Deployment failed."""


class AsyncClient:
    def __init__(
        self,
        client: client,
        loop: AbstractEventLoop | None = None,
        executor: ThreadPoolExecutor | None = None,
    ):
        self._client = client
        self._loop = loop or get_running_loop()
        self._executor = executor or ThreadPoolExecutor()

    def __getattr__(self, name: str) -> Callable[..., any]:
        attr = getattr(self._client, name)
        if not callable(attr):
            return attr

        async def fn(*args, **kwargs):
            return await self._loop.run_in_executor(
                self._executor, lambda: attr(*args, **kwargs)
            )

        return fn


class Aws:
    _ec2: AsyncClient
    _rds: AsyncClient
    _session: Session

    def __init__(
        self, settings: dict | None = None, token_pair: str | None = None
    ):
        if settings and "token_pair" in settings and not token_pair:
            token_pair = settings.pop("token_pair")

        aws_settings = {}

        if token_pair:
            access_key, secret_key = token_pair.split(":", 1)
            aws_settings.update(
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
            )

        if settings and "zone" in settings:
            aws_settings["region_name"] = settings["zone"]

        self._session = Session(**aws_settings)
        self._ec2 = AsyncClient(self._session.client("ec2"))
        self._rds = AsyncClient(self._session.client("rds"))

    async def get_vpc_id(self, name: str) -> str:
        response = await self._ec2.describe_vpcs(
            Filters=[{"Name": "tag:Name", "Values": [name]}]
        )
        vpcs = response.get("Vpcs", [])
        if not vpcs:
            raise DeployError(f"VPC {name!r} not found")
        return vpcs[0]["VpcId"]

    async def create_vpc_if_missing(self, name: str, cidr: str) -> str:
        """Return an existing VPC id or create a new VPC."""
        try:
            return await self.get_vpc_id(name)
        except DeployError:
            response = await self._ec2.create_vpc(CidrBlock=cidr)
            vpc_id = response["Vpc"]["VpcId"]
            await self._ec2.create_tags(
                Resources=[vpc_id], Tags=[{"Key": "Name", "Value": name}]
            )
            waiter = await self._ec2.get_waiter("vpc_available")
            waiter.wait(VpcIds=[vpc_id])
            return vpc_id

    async def get_security_group(self, name: str, vpc_id: str) -> str:
        response = await self._ec2.describe_security_groups(
            Filters=[{"Name": "group-name", "Values": [name]}]
        )
        groups = response.get("SecurityGroups", [])
        if groups:
            return groups[0]["GroupId"]

        response = await self._ec2.create_security_group(
            GroupName=name,
            Description="avalan deployment",
            VpcId=vpc_id,
        )
        return response["GroupId"]

    async def configure_security_group(self, group_id: str, port: int) -> None:
        try:
            await self._ec2.authorize_security_group_ingress(
                GroupId=group_id,
                IpProtocol="tcp",
                FromPort=port,
                ToPort=port,
                CidrIp="0.0.0.0/0",
            )
        except ClientError as exc:
            if exc.response["Error"]["Code"] != "InvalidPermission.Duplicate":
                raise

    async def create_rds_if_missing(
        self, db_id: str, instance_class: str, sg_id: str, storage: int
    ) -> str:
        try:
            await self._rds.describe_db_instances(DBInstanceIdentifier=db_id)
        except self._rds.exceptions.DBInstanceNotFoundFault:
            await self._rds.create_db_instance(
                DBInstanceIdentifier=db_id,
                DBInstanceClass=instance_class,
                Engine="postgres",
                MasterUsername="postgres",
                MasterUserPassword="postgres",
                AllocatedStorage=storage,
                VpcSecurityGroupIds=[sg_id],
                Tags=[{"Key": "Name", "Value": db_id}],
            )
            waiter = await self._rds.get_waiter("db_instance_available")
            waiter.wait(DBInstanceIdentifier=db_id)
        return db_id

    async def create_instance_if_missing(
        self,
        vpc_id: str,
        sg_id: str,
        ami_id: str,
        instance_type: str,
        instance_name: str,
        agent_path: str,
        port: int,
    ) -> str:
        user_data = self._create_user_data(agent_path, port)
        response = await self._ec2.describe_instances(
            Filters=[{"Name": "tag:Name", "Values": [instance_name]}]
        )
        reservations = response.get("Reservations", [])
        if reservations:
            return reservations[0]["Instances"][0]["InstanceId"]

        response = await self._ec2.describe_subnets(
            Filters=[{"Name": "vpc-id", "Values": [vpc_id]}]
        )

        subnet = response["Subnets"][0]["SubnetId"]

        response = await self._ec2.run_instances(
            ImageId=ami_id,
            InstanceType=instance_type,
            MinCount=1,
            MaxCount=1,
            SecurityGroupIds=[sg_id],
            SubnetId=subnet,
            UserData=user_data,
            TagSpecifications=[
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": "Name", "Value": instance_name}],
                }
            ],
        )
        return response["Instances"][0]["InstanceId"]

    def _create_user_data(self, agent_path: str, port: int) -> str:
        cmd = f"avalan agent serve {agent_path} --host 0.0.0.0 --port {port}\n"
        service = f"""
    [Unit]
    Description=Avalan Agent
    After=network.target

    [Service]
    Type=simple
    ExecStart=/usr/local/bin/{cmd}
    Restart=always

    [Install]
    WantedBy=multi-user.target
    """
        script = (
            "#!/bin/bash\n"
            "apt-get update -y\n"
            "apt-get install -y python3-pip\n"
            "pip3 install avalan\n"
            f"echo '{service}' > /etc/systemd/system/avalan.service\n"
            "systemctl daemon-reload\n"
            "systemctl enable avalan.service\n"
            "systemctl start avalan.service\n"
        )
        return script



from . import Tool, ToolSet
from ..compat import override
from ..entities import ToolCallContext
from ..memory.manager import MemoryManager
from ..memory.permanent import VectorFunction
from contextlib import AsyncExitStack


class MessageReadTool(Tool):
    """Fetch user information and preferences from previous messages.

    Args:
        search: What information to fetch. For example: "user's name".

    Returns:
        The search results, if any. This might be part of a message,
        so extract the information from it. For example, you may be
        getting "my name is Leo", not simply "Leo" as the user's name.
        If no result returned (indicated by the string "NOT_FOUND"), the
        information could not be found, so you should ask for it.
    """

    _memory_manager: MemoryManager
    _NOT_FOUND = "NOT_FOUND"

    def __init__(self, memory_manager: MemoryManager) -> None:
        super().__init__()
        self._memory_manager = memory_manager
        self.__name__ = "message.read"

    async def __call__(self, search: str, context: ToolCallContext) -> str:
        if (
            not context.agent_id
            or not context.session_id
            or not context.participant_id
        ):
            return MessageReadTool._NOT_FOUND

        results = await self._memory_manager.search_messages(
            agent_id=context.agent_id,
            exclude_session_id=context.session_id,
            function=VectorFunction.L2_DISTANCE,
            participant_id=context.participant_id,
            search=search,
            search_user_messages=True,
            limit=1,
        )
        if results and results[0].message:
            return results[0].message.content

        return MessageReadTool._NOT_FOUND


class MemoryToolSet(ToolSet):
    @override
    def __init__(
        self,
        memory_manager: MemoryManager,
        *,
        exit_stack: AsyncExitStack | None = None,
        namespace: str | None = None,
    ) -> None:
        tools = [MessageReadTool(memory_manager)]
        super().__init__(
            exit_stack=exit_stack, namespace=namespace, tools=tools
        )



from datetime import datetime, timezone
from logging import Logger
from typing import Any
from uuid import UUID, uuid4

from elasticsearch import AsyncElasticsearch

from ....memory.partitioner.text import TextPartition
from ....memory.permanent import (
    Memory,
    MemoryType,
    PermanentMemory,
    VectorFunction,
)
from . import ElasticsearchMemory, to_thread  # noqa: F401


class ElasticsearchRawMemory(ElasticsearchMemory, PermanentMemory):
    _index: str
    _client: Any
    _logger: Logger

    def __init__(
        self,
        index: str,
        *,
        client: Any,
        logger: Logger,
    ) -> None:
        ElasticsearchMemory.__init__(
            self, index=index, client=client, logger=logger
        )
        PermanentMemory.__init__(self, sentence_model=None)

    @classmethod
    async def create_instance(
        cls,
        index: str,
        *,
        logger: Logger,
        es_client: Any | None = None,
    ) -> "ElasticsearchRawMemory":
        if es_client is None:
            es_client = AsyncElasticsearch()
        memory = cls(index=index, client=es_client, logger=logger)
        return memory

    async def append_with_partitions(
        self,
        namespace: str,
        participant_id: UUID,
        *,
        memory_type: MemoryType,
        data: str,
        identifier: str,
        partitions: list[TextPartition],
        symbols: dict | None = None,
        model_id: str | None = None,
    ) -> None:
        assert (
            namespace and participant_id and data and identifier and partitions
        )
        now_utc = datetime.now(timezone.utc)
        entry, partition_rows = self._build_memory_with_partitions(
            namespace,
            participant_id,
            memory_type,
            data,
            identifier,
            partitions,
            created_at=now_utc,
            symbols=symbols,
            model_id=model_id,
            memory_id=uuid4(),
        )
        await self._index_document(
            index=self._index,
            id=str(entry.id),
            document={
                "id": str(entry.id),
                "model_id": entry.model_id,
                "type": str(entry.type),
                "participant_id": str(entry.participant_id),
                "namespace": entry.namespace,
                "identifier": entry.identifier,
                "data": entry.data,
                "partitions": entry.partitions,
                "symbols": entry.symbols,
                "created_at": entry.created_at.isoformat(),
            },
        )
        for row in partition_rows:
            await self._index_vector(
                index=self._index,
                id=f"{row.memory_id}:{row.partition}",
                vector=row.embedding.tolist(),
                metadata={
                    "memory_id": str(row.memory_id),
                    "participant_id": str(row.participant_id),
                    "namespace": namespace,
                },
            )

    async def search_memories(
        self,
        *,
        search_partitions: list[TextPartition],
        participant_id: UUID,
        namespace: str,
        function: VectorFunction,
        limit: int | None = None,
    ) -> list[Memory]:
        assert participant_id and namespace and search_partitions
        query = search_partitions[0].embeddings.tolist()
        response = await self._query_vector(
            index=self._index,
            query_vector=query,
            top_k=limit or 10,
            function=str(function),
            filter={
                "memory_id": "*",
                "participant_id": str(participant_id),
                "namespace": namespace,
            },
        )
        results: list[Memory] = []
        for item in response.get("Items", []):
            mem_id = item.get("Metadata", {}).get("memory_id")
            if not mem_id:
                continue
            obj = await self._get_document(index=self._index, id=mem_id)
            meta = obj.get("_source") if obj else None
            if not meta:
                continue
            results.append(
                Memory(
                    id=UUID(meta["id"]),
                    model_id=meta["model_id"],
                    type=MemoryType(meta["type"]),
                    participant_id=UUID(meta["participant_id"]),
                    namespace=meta["namespace"],
                    identifier=meta["identifier"],
                    data=meta["data"],
                    partitions=meta["partitions"],
                    symbols=meta["symbols"],
                    created_at=datetime.fromisoformat(meta["created_at"]),
                )
            )
        return results

    async def search(self, query: str) -> list[Memory] | None:
        raise NotImplementedError()



from ....agent import EngineEnvironment, AgentOperation, Goal, Specification
from ....agent.orchestrator import Orchestrator
from ....entities import EngineUri, Modality, TransformerEngineSettings
from ....event.manager import EventManager
from ....memory.manager import MemoryManager
from ....model.manager import ModelManager
from ....tool.manager import ToolManager
from logging import Logger
from uuid import UUID


class DefaultOrchestrator(Orchestrator):
    def __init__(
        self,
        engine_uri: EngineUri,
        logger: Logger,
        model_manager: ModelManager,
        memory: MemoryManager,
        tool: ToolManager,
        event_manager: EventManager,
        *,
        name: str | None,
        role: str | None,
        task: str | None,
        instructions: str | None,
        rules: list[str] | None,
        template_id: str | None = None,
        settings: TransformerEngineSettings | None = None,
        call_options: dict | None = None,
        template_vars: dict | None = None,
        id: UUID | None = None,
    ):
        specification = Specification(
            role=role,
            goal=(
                Goal(task=task, instructions=[instructions])
                if task and instructions
                else None
            ),
            rules=rules,
            template_id=template_id or "agent.md",
            template_vars=template_vars,
        )
        super().__init__(
            logger,
            model_manager,
            memory,
            tool,
            event_manager,
            AgentOperation(
                specification=specification,
                environment=EngineEnvironment(
                    engine_uri=engine_uri, settings=settings
                ),
                modality=Modality.TEXT_GENERATION,
            ),
            call_options=call_options,
            id=id,
            name=name,
        )



from . import Tool, ToolSet
from ..compat import override
from ..entities import Message, MessageRole, ToolCallContext
from ..filters import Partitioner
from contextlib import AsyncExitStack
from dataclasses import dataclass
from email.message import EmailMessage
from faiss import IndexFlatL2
from io import BytesIO, TextIOBase
from markitdown import MarkItDown
from numpy import vstack
from playwright.async_api import (
    async_playwright,
    Browser,
    Page,
    PlaywrightContextManager,
)
from typing import Literal


@dataclass(frozen=True, kw_only=True)
class BrowserToolSettings(dict):
    engine: Literal["chromium", "firefox", "webkit"] = "firefox"
    search: bool = False
    search_context: int | None = 10
    search_k: int = 1
    debug: bool = False
    debug_url: str | None = None
    debug_source: TextIOBase | None = None
    slowdown: int | None = None
    devtools: bool = False
    chromium_sandbox: bool = True
    viewport_width: int = 1024
    viewport_height: int = 768
    scale_factor: float = 1.0
    is_mobile: bool = False
    has_touch: bool = False
    java_script_enabled: bool = True

    def __post_init__(self):
        self["debug"] = self.debug
        self["debug_url"] = self.debug_url
        self["debug_source"] = self.debug_source
        self["engine"] = self.engine
        self["search"] = self.search
        self["search_context"] = self.search_context
        self["search_k"] = self.search_k
        self["slowdown"] = self.slowdown
        self["devtools"] = self.devtools
        self["chromium_sandbox"] = self.chromium_sandbox
        self["viewport_width"] = self.viewport_width
        self["viewport_height"] = self.viewport_height
        self["scale_factor"] = self.scale_factor
        self["is_mobile"] = self.is_mobile
        self["has_touch"] = self.has_touch
        self["java_script_enabled"] = self.java_script_enabled


class BrowserTool(Tool):
    """
    Uses a web browser to access the internet and get the contents of a URL.

    Args:
        url: URL of a page to open.

    Returns:
        Contents of the given URL page.
    """

    _client: PlaywrightContextManager
    _settings: BrowserToolSettings
    _browser: Browser | None = None
    _page: Page | None = None
    _md: MarkItDown | None = None
    _partitioner: Partitioner | None = None

    def __init__(
        self,
        settings: BrowserToolSettings,
        client: PlaywrightContextManager,
        partitioner: Partitioner | None = None,
    ) -> None:
        super().__init__()
        self._settings = settings
        self._client = client
        self._md = MarkItDown()
        self._partitioner = partitioner
        self.__name__ = "open"

    async def __call__(self, url: str, *, context: ToolCallContext) -> str:
        content = await self._read(url)

        if (
            self._settings.search
            and self._partitioner
            and context.input
            and isinstance(context.input, Message)
            and context.input.role == MessageRole.USER
        ):
            query = context.input.content
            sentence_model = self._partitioner.sentence_model
            knowledge_partitions = await self._partitioner(content)

            knowledge_embeddings = [
                kp.embeddings for kp in knowledge_partitions
            ]
            query_embeddings = await sentence_model([query])

            assert knowledge_embeddings and query_embeddings.any()

            knowledge_stack = vstack(knowledge_embeddings).astype(
                "float32", copy=False
            )
            query_stack = vstack(query_embeddings).astype(
                "float32", copy=False
            )

            index = IndexFlatL2(knowledge_embeddings[0].shape[0])
            index.add(knowledge_stack)

            distances, ids = index.search(query_stack, self._settings.search_k)

            matches: list[tuple[int, int, float]] = [
                (q_id, kn_id, float(dist))
                for q_id, (dist_row, id_row) in enumerate(zip(distances, ids))
                for dist, kn_id in zip(dist_row, id_row)
            ]
            # smallest distance first
            matches.sort(key=lambda t: t[2])

            for q_id, kn_id, l2_distance in matches:
                knowledge_chunk = (
                    knowledge_partitions[kn_id].data
                    if knowledge_partitions
                    else query if kn_id == 0 else None
                )
                if not knowledge_chunk:
                    continue

                knowledge_match = (
                    "\n".join(
                        [
                            kp.data
                            for kp in knowledge_partitions[
                                max(
                                    kn_id - self._settings.search_context, 0
                                ) : min(
                                    kn_id + self._settings.search_context + 1,
                                    len(knowledge_partitions),
                                )
                            ]
                        ]
                    )
                    if self._settings.search_context
                    else knowledge_chunk
                )

                content = knowledge_match

        return content

    async def _read(self, url: str) -> str:
        if (
            self._settings.debug
            and self._settings.debug_url
            and url == self._settings.debug_url
            and self._settings.debug_source
        ):
            assert isinstance(self._settings.debug_source, TextIOBase)
            content = self._settings.debug_source.read()
            return content

        if not self._browser:
            browser_type = (
                self._client.chromium
                if self._settings.engine == "chromium"
                else (
                    self._client.firefox
                    if self._settings.engine == "firefox"
                    else (
                        self._client.webkit
                        if self._settings.engine == "webkit"
                        else None
                    )
                )
            )
            assert browser_type
            self._browser = await browser_type.launch(
                headless=True,
                slow_mo=self._settings.slowdown,
                devtools=self._settings.devtools,
                executable_path=None,
                args=[],
                env={},
                timeout=0,
                firefox_user_prefs={},
                chromium_sandbox=self._settings.chromium_sandbox,
            )

        if not self._page:
            self._page = await self._browser.new_page(
                viewport={
                    "width": self._settings.viewport_width,
                    "height": self._settings.viewport_height,
                },
                screen=None,
                device_scale_factor=self._settings.scale_factor,
                is_mobile=self._settings.is_mobile,
                has_touch=self._settings.has_touch,
                java_script_enabled=self._settings.java_script_enabled,
                locale=None,
                timezone_id=None,
                geolocation=None,
                permissions=[],
                extra_http_headers={},
                ignore_https_errors=False,
                bypass_csp=False,
                offline=False,
                http_credentials=None,
                storage_state=None,
                accept_downloads=True,
                base_url=None,
                color_scheme=None,
            )

        response = await self._page.goto(url)
        contents: str = await self._page.content()
        content_type_header = response.headers.get("content-type", None)
        assert content_type_header

        m = EmailMessage()
        m["content-type"] = content_type_header
        maintype = m.get_content_maintype() or "text"
        assert maintype == "text"
        encoding = (m.get_param("charset") or "utf-8").lower()
        mime_type = m.get_content_type()
        byte_stream = BytesIO(contents.encode(encoding))
        result = self._md.convert_stream(byte_stream, mime_type=mime_type)
        content = result.text_content
        return content

    def with_client(self, client: PlaywrightContextManager) -> "BrowserTool":
        self._client = client
        return self

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: BaseException | None,
    ) -> bool:
        if self._page:
            await self._page.close()
        if self._browser:
            await self._browser.close()

        return await super().__aexit__(exc_type, exc_value, traceback)


class BrowserToolSet(ToolSet):
    _client: PlaywrightContextManager | None = None

    @override
    def __init__(
        self,
        settings: BrowserToolSettings,
        *,
        exit_stack: AsyncExitStack | None = None,
        namespace: str | None = None,
        partitioner: Partitioner | None = None,
    ):
        assert settings

        if not exit_stack:
            exit_stack = AsyncExitStack()

        self._client = async_playwright()

        tools = [BrowserTool(settings, self._client, partitioner=partitioner)]
        return super().__init__(
            exit_stack=exit_stack, namespace=namespace, tools=tools
        )

    @override
    async def __aenter__(self) -> "BrowserToolSet":
        self._client = await self._exit_stack.enter_async_context(self._client)
        for i, tool in enumerate(self._tools):
            self._tools[i] = tool.with_client(self._client)
        return await super().__aenter__()



from asyncio import sleep
from ....compat import override
from ....entities import (
    GenerationSettings,
    Input,
    Message,
    MessageRole,
    ProbabilityDistribution,
    TextGenerationLoaderClass,
    Token,
    TokenDetail,
    TransformerEngineSettings,
)
from ....model.response.text import TextGenerationResponse
from ....model.vendor import TextGenerationVendor
from ....model.nlp import BaseNLPModel
from ....model.engine import Engine
from diffusers import DiffusionPipeline
from ....tool.manager import ToolManager
from dataclasses import asdict, replace
from importlib.util import find_spec
from logging import Logger, getLogger
from threading import Thread
from torch import log_softmax, softmax, topk, Tensor
from torch.nn.functional import gumbel_softmax
from transformers import (
    AsyncTextIteratorStreamer,
    AutoModelForCausalLM,
    Gemma3ForConditionalGeneration,
    Mistral3ForConditionalGeneration,
    PreTrainedModel,
)
from transformers.generation import StoppingCriteria
from transformers.tokenization_utils_base import BatchEncoding
from typing import AsyncGenerator, Literal


class TextGenerationModel(BaseNLPModel):
    _loaders: dict[TextGenerationLoaderClass, type[PreTrainedModel]] = {
        "auto": AutoModelForCausalLM,
        "gemma3": Gemma3ForConditionalGeneration,
        "mistral3": Mistral3ForConditionalGeneration,
    }

    def __init__(
        self,
        model_id: str,
        settings: TransformerEngineSettings | None = None,
        logger: Logger = getLogger(__name__),
    ) -> None:
        super().__init__(
            model_id, settings or TransformerEngineSettings(), logger
        )

    @property
    def supports_sample_generation(self) -> bool:
        return True

    @property
    def supports_token_streaming(self) -> bool:
        return True

    def _load_model(
        self,
    ) -> PreTrainedModel | TextGenerationVendor | DiffusionPipeline:
        assert (
            self._settings.loader_class in self._loaders
        ), f"Unrecognized loader {self._settings.loader_class}"

        if self._settings.quantization and find_spec("bitsandbytes"):
            from transformers import BitsAndBytesConfig

            quantization = self._settings.quantization
            bnb_config = BitsAndBytesConfig(
                load_in_4bit=quantization.load_in_4bit,
                bnb_4bit_quant_type=quantization.bnb_4bit_quant_type,
                bnb_4bit_use_double_quant=quantization.bnb_4bit_use_double_quant,
                bnb_4bit_compute_dtype=quantization.bnb_4bit_compute_dtype,
            )
        else:
            bnb_config = None

        loader = self._loaders[self._settings.loader_class]
        model = loader.from_pretrained(
            self._model_id,
            cache_dir=self._settings.cache_dir,
            subfolder=self._settings.subfolder or "",
            attn_implementation=self._settings.attention,
            output_hidden_states=self._settings.output_hidden_states,
            trust_remote_code=self._settings.trust_remote_code,
            state_dict=self._settings.state_dict,
            local_files_only=self._settings.local_files_only,
            low_cpu_mem_usage=(
                True if self._device else self._settings.low_cpu_mem_usage
            ),
            torch_dtype=Engine.weight(self._settings.weight_type),
            device_map=self._device,
            token=self._settings.access_token,
            quantization_config=bnb_config,
            revision=self._settings.revision,
            tp_plan=Engine._get_tp_plan(self._settings.parallel),
        )
        return model

    @override
    async def __call__(
        self,
        input: Input,
        system_prompt: str | None = None,
        settings: GenerationSettings | None = None,
        stopping_criterias: list[StoppingCriteria] | None = None,
        *,
        manual_sampling: bool = False,
        pick: int | None = None,
        skip_special_tokens: bool = False,
        tool: ToolManager | None = None,
    ) -> TextGenerationResponse:
        assert self._tokenizer, (
            f"Model {self._model} can't be executed "
            + "without a tokenizer loaded first"
        )
        assert self._model, (
            f"Model {self._model} can't be executed, it "
            + "needs to be loaded first"
        )

        if not settings:
            settings = GenerationSettings()
        assert settings.temperature is None or (
            settings.temperature > 0 or settings.temperature == 0.0
        ), (
            "Temperature has to be a strictly positive float or zero, "
            + "otherwise your next token scores will be invalid"
        )

        do_sample = (
            settings.do_sample if self.supports_sample_generation else False
        )
        if self.supports_sample_generation and settings.temperature:
            do_sample = True

        assert (not do_sample and not settings.temperature) or (
            do_sample and settings.temperature
        ), "Sample-based generation can only be set with temperature"

        output_fn = (
            self._string_output
            if not settings.use_async_generator
            else (
                self._token_generator
                if manual_sampling
                else self._stream_generator
            )
        )
        generation_settings = replace(
            settings,
            do_sample=do_sample,
            pad_token_id=(
                settings.pad_token_id
                if settings.pad_token_id is not None
                else self._tokenizer.eos_token_id
            ),
        )
        inputs = self._tokenize_input(
            input,
            system_prompt,
            context=None,
            tool=tool,
            chat_template_settings=asdict(settings.chat_settings),
        )
        return TextGenerationResponse(
            output_fn,
            inputs=inputs,
            logger=self._logger,
            generation_settings=generation_settings,
            pick=pick,
            settings=generation_settings,
            stopping_criterias=stopping_criterias,
            skip_special_tokens=skip_special_tokens,
            use_async_generator=settings.use_async_generator,
        )

    async def _stream_generator(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        stopping_criterias: list[StoppingCriteria] | None,
        skip_special_tokens: bool,
        **kwargs,
    ) -> AsyncGenerator[str, None]:
        _l = self._log

        streamer = AsyncTextIteratorStreamer(
            self._tokenizer,
            skip_prompt=True,
            decode_kwargs={"skip_special_tokens": skip_special_tokens},
        )

        _l("Created generator async text token streamer")

        def generate_stream() -> None:
            _l(
                f"Streaming up to {settings.max_new_tokens} tokens "
                f"{'with' if settings.do_sample else 'without'} sample "
                f"and {settings.temperature} temperature"
            )
            self._generate_output(
                inputs,
                settings,
                stopping_criterias,
                streamer=streamer,
            )

        thread = Thread(
            target=generate_stream, name=f"{self._model_id}/generate_stream"
        )
        thread.start()

        _l(f"Generation thread #{thread.ident} ({thread.name}) started")

        async for token in streamer:
            yield token

        thread.join()

        _l(f"Generation thread #{thread.ident} ({thread.name}) finished")

    def _string_output(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        stopping_criterias: list[StoppingCriteria] | None,
        skip_special_tokens: bool,
        **kwargs,
    ) -> str:
        _l = self._log
        input_length = inputs["input_ids"].shape[1]
        outputs = self._generate_output(inputs, settings, stopping_criterias)
        response = self._tokenizer.decode(
            outputs[0][input_length:], skip_special_tokens=False
        )
        return response

    async def _token_generator(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        stopping_criterias: list[StoppingCriteria] | None,
        skip_special_tokens: bool,
        pick: int | None,
        probability_distribution: ProbabilityDistribution = "softmax",
    ) -> AsyncGenerator[Token | TokenDetail, None]:
        assert not settings.temperature or (
            settings.temperature >= 0 and settings.temperature <= 1
        ), "temperature should be [0, 1]"
        assert not pick or pick >= 0

        _l = self._log

        enable_entmax = find_spec("entmax") and probability_distribution in [
            "entmax",
            "sparsemax",
        ]
        if enable_entmax:
            import entmax

        _l(
            f"Generating up to {settings.max_new_tokens} tokens "
            f"{'with' if settings.do_sample else 'without'} sample "
            f"and {settings.temperature} temperature"
        )

        generation_settings = replace(
            settings,
            return_dict_in_generate=True,
            output_scores=True,
        )
        outputs = self._generate_output(
            inputs, generation_settings, stopping_criterias
        )
        sequences = outputs.sequences[0]
        scores = outputs.scores  # list of logits for each generated token
        start = inputs["input_ids"].shape[1]  # where generation began
        generated_sequences = sequences[start:]

        _l(f"Generated {len(generated_sequences)} sequences")

        total_tokens = 0
        for step, token_id in enumerate(generated_sequences):
            _l(f"Got step {step} token {token_id}")

            # logits are the raw-unnormalized scores output by the final
            # linear layer
            tensor = scores[step]  # scores is (batch_size, vocab_size)
            logits = tensor[0]  # first element in batch dimension

            # apply probabilty  distribution over last tensor layer, vocab_size
            logits_probs = (
                log_softmax(logits, dim=-1)
                if probability_distribution == "log_softmax"
                else (
                    gumbel_softmax(
                        logits, tau=settings.temperature, hard=False, dim=-1
                    )
                    if probability_distribution == "gumbel_softmax"
                    else (
                        entmax.sparsemax(logits, dim=-1)
                        if enable_entmax
                        and probability_distribution == "sparsemax"
                        else (
                            entmax.entmax15(logits, dim=-1)
                            if enable_entmax
                            and probability_distribution == "entmax"
                            else softmax(logits / settings.temperature, dim=-1)
                        )
                    )
                )
            )

            tokens: list[Token] | None = None
            if pick > 0:
                picked_logits = topk(logits_probs, pick)
                picked_logits_ids = picked_logits.indices.tolist()
                picked_logits_probs = picked_logits.values.tolist()
                tokens = [
                    Token(
                        id=token_id,
                        token=self._tokenizer.decode(
                            token_id, skip_special_tokens=skip_special_tokens
                        ),
                        probability=picked_logits_probs[i],
                    )
                    for i, token_id in enumerate(picked_logits_ids)
                ]

            raw_token = TokenDetail(
                id=token_id,
                token=self._tokenizer.decode(
                    token_id, skip_special_tokens=skip_special_tokens
                ),
                probability=logits_probs[token_id].item(),
                step=step,
                probability_distribution=probability_distribution,
                tokens=tokens,
            )

            _l(f"Yielding step {step} token detail {raw_token.__repr__()}")

            yield raw_token

            total_tokens = total_tokens + 1

        _l(f"Yielded {total_tokens}")

        await sleep(0)  # and just like that, a generator is an async generator

    def _tokenize_input(
        self,
        input: Input,
        system_prompt: str | None,
        context: str | None,
        tensor_format: Literal["pt"] = "pt",
        chat_template: str | None = None,
        chat_template_settings: dict[str, object] | None = None,
        tool: ToolManager | None = None,
    ) -> dict[str, Tensor] | BatchEncoding | Tensor:
        _l = self._log
        messages = self._messages(input, system_prompt, tool)
        template_messages = [
            {"role": message.role, "content": message.content}
            for message in messages
        ]

        if not self._tokenizer.chat_template:
            _l("Model does not support template messages, so staying plain")

            prompt = f"{system_prompt}\n\n" or ""
            use_prefix = not any(
                tm["role"] == MessageRole.USER for tm in template_messages
            )

            for template_message in template_messages:
                if use_prefix:
                    prompt += (
                        "User: "
                        if template_message["role"] == MessageRole.USER
                        else (
                            "Assistant: "
                            if template_message["role"]
                            == MessageRole.ASSISTANT
                            else ""
                        )
                    )
                prompt += template_message["content"].strip() + "\n"

            inputs = self._tokenizer(
                prompt, add_special_tokens=True, return_tensors=tensor_format
            )
        else:
            _l(f"Got {len(template_messages)} template messages")

            _l(f"Applying chat template to {len(template_messages)} messages")
            inputs = self._tokenizer.apply_chat_template(
                template_messages,
                chat_template=chat_template,
                tools=tool.json_schemas() if tool else None,
                **(chat_template_settings or {}),
                return_tensors=tensor_format,
            )

        if hasattr(self._model, "device"):
            _l(f"Translating inputs to {self._model.device}")
            inputs = inputs.to(self._model.device)
        return inputs

    def _messages(
        self,
        input: Input,
        system_prompt: str | None,
        tool: ToolManager | None = None,
    ) -> list[Message]:
        if isinstance(input, str):
            input = Message(role=MessageRole.USER, content=input)
        elif isinstance(input, list):
            for m in input:
                assert isinstance(m, Message)
        elif not isinstance(input, Message):
            raise ValueError(input)

        messages = [input] if not isinstance(input, list) else input

        if system_prompt:
            messages = [
                Message(role=MessageRole.SYSTEM, content=system_prompt)
            ] + messages

        assert isinstance(messages, list)
        return messages



from abc import ABC
from ...entities import GenerationSettings, WeightType as WeightType
from ...model.transformer import TransformerModel
from contextlib import nullcontext
from torch import (
    inference_mode,
    Tensor,
)
from transformers import AsyncTextIteratorStreamer
from transformers.generation import StoppingCriteria
from transformers.tokenization_utils_base import BatchEncoding


class BaseNLPModel(TransformerModel, ABC):
    def _generate_output(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        stopping_criterias: list[StoppingCriteria] | None = None,
        streamer: AsyncTextIteratorStreamer | None = None,
    ):
        eos_token_id = (
            settings.eos_token_id
            if settings.eos_token_id
            else (
                self._tokenizer.eos_token_id
                if not settings.forced_eos_token_id and self._tokenizer
                else None
            )
        )
        generation_kwargs = {
            "bos_token_id": settings.bos_token_id,
            "diversity_penalty": settings.diversity_penalty,
            "do_sample": settings.do_sample,
            "early_stopping": settings.early_stopping,
            "eos_token_id": eos_token_id,
            "forced_bos_token_id": settings.forced_bos_token_id,
            "forced_eos_token_id": settings.forced_eos_token_id,
            "max_length": settings.max_length,
            "max_new_tokens": settings.max_new_tokens,
            "max_time": settings.max_time,
            "min_length": settings.min_length,
            "min_new_tokens": settings.min_new_tokens,
            "min_p": settings.min_p,
            "num_beams": settings.num_beams,
            "num_beam_groups": settings.num_beam_groups,
            "num_return_sequences": settings.num_return_sequences,
            "output_attentions": settings.output_attentions,
            "output_hidden_states": settings.output_hidden_states,
            "output_logits": settings.output_logits,
            "output_scores": settings.output_scores,
            "pad_token_id": (
                settings.pad_token_id or self._tokenizer.eos_token_id
            ),
            "penalty_alpha": settings.penalty_alpha,
            "prompt_lookup_num_tokens": settings.prompt_lookup_num_tokens,
            "repetition_penalty": settings.repetition_penalty,
            "return_dict_in_generate": settings.return_dict_in_generate,
            "stop_strings": settings.stop_strings,
            "stopping_criteria": stopping_criterias,
            "streamer": streamer,
            "temperature": settings.temperature,
            "top_k": settings.top_k,
            "top_p": settings.top_p,
            "use_cache": settings.use_cache,
        }

        attention_mask: Tensor | None = None
        if settings.use_inputs_attention_mask:
            attention_mask = (
                inputs.get("attention_mask", None)
                if isinstance(inputs, BatchEncoding)
                else getattr(inputs, "attention_mask", None)
            )
            if attention_mask is not None:
                assert isinstance(attention_mask, Tensor)
                assert attention_mask.shape == inputs["input_ids"].shape
                generation_kwargs["attention_mask"] = attention_mask

        if (
            not settings.use_inputs_attention_mask
            or attention_mask is not None
        ):
            inputs.pop("attention_mask", None)

        if settings.forced_bos_token_id or settings.forced_eos_token_id:
            del generation_kwargs["bos_token_id"]
            del generation_kwargs["eos_token_id"]

        with (
            inference_mode()
            if not settings.enable_gradient_calculation
            else nullcontext()
        ):
            outputs = (
                self._model.generate(
                    inputs, tokenizer=self._tokenizer, **generation_kwargs
                )
                if isinstance(inputs, Tensor)
                else self._model.generate(
                    **inputs, tokenizer=self._tokenizer, **generation_kwargs
                )
            )
        return outputs



from boto3 import client
from . import Secrets


class AwsSecrets(Secrets):
    _SERVICE = "secretsmanager"

    def __init__(self, aws_client: object | None = None):
        self._client = aws_client or client(self._SERVICE)

    def read(self, key: str) -> str | None:
        response = self._client.get_secret_value(SecretId=key)
        return response.get("SecretString")

    def write(self, key: str, secret: str) -> None:
        self._client.put_secret_value(SecretId=key, SecretString=secret)

    def delete(self, key: str) -> None:
        self._client.delete_secret(
            SecretId=key,
            ForceDeleteWithoutRecovery=True,
        )



from ..flow.connection import Connection
from ..flow.node import Node
from typing import Any, Callable
from re import match


class Flow:
    def __init__(self) -> None:
        self.nodes: dict[str, Node] = {}
        self.connections: list[Connection] = []
        self.outgoing: dict[str, list[Connection]] = {}

    def add_node(self, node: Node) -> None:
        self.nodes[node.name] = node
        self.outgoing.setdefault(node.name, [])

    def add_connection(
        self,
        src_name: str,
        dest_name: str,
        label: str | None = None,
        conditions: list[Callable[[Any], bool]] | None = None,
        filters: list[Callable[[Any], Any]] | None = None,
    ) -> None:
        if src_name not in self.nodes or dest_name not in self.nodes:
            raise KeyError(f"Unknown node: {src_name} or {dest_name}")
        src = self.nodes[src_name]
        dest = self.nodes[dest_name]
        conn = Connection(src, dest, label, conditions, filters)
        self.connections.append(conn)
        self.outgoing[src_name].append(conn)

    def parse_mermaid(self, mermaid: str) -> None:
        lines = [line.strip() for line in mermaid.splitlines() if line.strip()]
        if lines and lines[0].lower().startswith("graph"):
            lines = lines[1:]
        for line in lines:
            if "-->" not in line:
                continue
            left, right = line.split("-->", 1)
            left = left.strip()
            right = right.strip()
            # strip edge label syntax: A -- text --> B
            label = None
            if "--" in left:
                parts = left.split("--", 1)
                left, label = parts[0].strip(), parts[1].strip()
            src_id, src_lbl, src_shape = self._parse_node(left)
            dst_id, dst_lbl, dst_shape = self._parse_node(right)
            for nid, nlbl, nshape in [
                (src_id, src_lbl, src_shape),
                (dst_id, dst_lbl, dst_shape),
            ]:
                if nid not in self.nodes:
                    self.add_node(Node(nid, label=nlbl or nid, shape=nshape))
                else:
                    node = self.nodes[nid]
                    if nlbl and node.label == node.name:
                        node.label = nlbl
                    if nshape and not node.shape:
                        node.shape = nshape
            self.add_connection(src_id, dst_id, label=label)

    def _parse_node(self, text: str) -> tuple[str, str | None, str | None]:
        m = match(r"^([A-Za-z0-9_]+)", text)
        if not m:
            return text, None, None
        nid = m.group(1)
        rem = text[len(nid) :].strip()
        if not rem:
            return nid, None, None
        shape = None
        label = None
        if rem.startswith("[") and rem.endswith("]"):
            shape = "rect"
            label = rem[1:-1]
        elif rem.startswith("(((") and rem.endswith(")))"):
            shape = "circle"
            label = rem[3:-3]
        elif rem.startswith("(") and rem.endswith(")"):
            shape = "roundrect"
            label = rem[1:-1]
        elif rem.startswith("{") and rem.endswith("}"):
            shape = "diamond"
            label = rem[1:-1]
        else:
            label = rem
        return nid, label, shape

    def execute(
        self,
        initial_node: str | Node | None = None,
        initial_data: Any = None,
    ) -> Any:
        # Determine start nodes
        if initial_node is None:
            indegree = {n: 0 for n in self.nodes}
            for c in self.connections:
                indegree[c.dest.name] += 1
            start = [self.nodes[n] for n, d in indegree.items() if d == 0]
        else:
            start = (
                [self.nodes[initial_node]]
                if isinstance(initial_node, str)
                else [initial_node]
            )  # type: ignore
        indegree_map = {n: 0 for n in self.nodes}
        for c in self.connections:
            indegree_map[c.dest.name] += 1

        buffers: dict[str, dict[str, Any]] = {n: {} for n in self.nodes}
        if initial_data is not None and len(start) == 1:
            buffers[start[0].name] = {"__init__": initial_data}

        ready: list[Node] = []
        for n in start:
            indegree_map[n.name] = 0
            ready.append(n)

        outputs: dict[str, Any] = {}
        while ready:
            node = ready.pop(0)
            inp = buffers[node.name]
            incoming = sum(
                1 for c in self.connections if c.dest.name == node.name
            )
            if incoming > 0 and not inp:
                outputs[node.name] = None
            else:
                outputs[node.name] = node.execute(inp)
            outval = outputs[node.name]
            if outval is not None:
                for c in self.outgoing.get(node.name, []):
                    if c.check_conditions(outval):
                        forwarded = c.apply_filters(outval)
                        buffers[c.dest.name][node.name] = forwarded
            for c in self.outgoing.get(node.name, []):
                indegree_map[c.dest.name] -= 1
                if indegree_map[c.dest.name] == 0:
                    ready.append(c.dest)

        terminal = {
            n: outputs[n] for n, outs in self.outgoing.items() if not outs
        }
        if len(terminal) == 1:
            return next(iter(terminal.values()))
        return terminal



from ....deploy.aws import AsyncClient
from ....memory.partitioner.text import TextPartition
from ....memory.permanent import (
    Memory,
    MemoryType,
    PermanentMemory,
    VectorFunction,
)
from . import S3VectorsMemory
from asyncio import to_thread  # noqa: F401
from datetime import datetime, timezone
from json import dumps, loads
from logging import Logger
from typing import Any
from uuid import UUID, uuid4

from boto3 import client as boto_client


class S3VectorsRawMemory(S3VectorsMemory, PermanentMemory):
    _bucket: str
    _collection: str
    _client: Any
    _logger: Logger

    def __init__(
        self,
        bucket: str,
        collection: str,
        *,
        client: Any,
        logger: Logger,
    ) -> None:
        S3VectorsMemory.__init__(
            self,
            bucket=bucket,
            collection=collection,
            client=client,
            logger=logger,
        )
        PermanentMemory.__init__(self, sentence_model=None)

    @classmethod
    async def create_instance(
        cls,
        bucket: str,
        collection: str,
        *,
        logger: Logger,
        aws_client: Any | None = None,
    ) -> "S3VectorsRawMemory":
        if aws_client is None:
            aws_client = boto_client("s3vectors")
        memory = cls(
            bucket=bucket,
            collection=collection,
            client=AsyncClient(aws_client),
            logger=logger,
        )
        return memory

    async def append_with_partitions(
        self,
        namespace: str,
        participant_id: UUID,
        *,
        memory_type: MemoryType,
        data: str,
        identifier: str,
        partitions: list[TextPartition],
        symbols: dict | None = None,
        model_id: str | None = None,
    ) -> None:
        assert (
            namespace and participant_id and data and identifier and partitions
        )
        now_utc = datetime.now(timezone.utc)
        entry, partition_rows = self._build_memory_with_partitions(
            namespace,
            participant_id,
            memory_type,
            data,
            identifier,
            partitions,
            created_at=now_utc,
            symbols=symbols,
            model_id=model_id,
            memory_id=uuid4(),
        )
        await self._put_object(
            Bucket=self._bucket,
            Key=f"{self._collection}/{entry.id}.json",
            Body=dumps(
                {
                    "id": str(entry.id),
                    "model_id": entry.model_id,
                    "type": str(entry.type),
                    "participant_id": str(entry.participant_id),
                    "namespace": entry.namespace,
                    "identifier": entry.identifier,
                    "data": entry.data,
                    "partitions": entry.partitions,
                    "symbols": entry.symbols,
                    "created_at": entry.created_at.isoformat(),
                }
            ).encode(),
        )
        for row in partition_rows:
            await self._put_vector(
                Bucket=self._bucket,
                Collection=self._collection,
                Id=f"{row.memory_id}:{row.partition}",
                Vector=row.embedding.tolist(),
                Metadata={
                    "memory_id": str(row.memory_id),
                    "participant_id": str(row.participant_id),
                    "namespace": namespace,
                },
            )

    async def search_memories(
        self,
        *,
        search_partitions: list[TextPartition],
        participant_id: UUID,
        namespace: str,
        function: VectorFunction,
        limit: int | None = None,
    ) -> list[Memory]:
        assert participant_id and namespace and search_partitions
        query = search_partitions[0].embeddings.tolist()
        response = await self._query_vector(
            Bucket=self._bucket,
            Collection=self._collection,
            QueryVector=query,
            TopK=limit or 10,
            Function=str(function),
            Filter={
                "memory_id": "*",
                "participant_id": str(participant_id),
                "namespace": namespace,
            },
        )
        results = []
        for item in response.get("Items", []):
            mem_id = item.get("Metadata", {}).get("memory_id")
            if not mem_id:
                continue
            obj = await self._get_object(
                Bucket=self._bucket,
                Key=f"{self._collection}/{mem_id}.json",
            )
            meta = loads(obj["Body"].read().decode())
            results.append(
                Memory(
                    id=UUID(meta["id"]),
                    model_id=meta["model_id"],
                    type=MemoryType(meta["type"]),
                    participant_id=UUID(meta["participant_id"]),
                    namespace=meta["namespace"],
                    identifier=meta["identifier"],
                    data=meta["data"],
                    partitions=meta["partitions"],
                    symbols=meta["symbols"],
                    created_at=datetime.fromisoformat(meta["created_at"]),
                )
            )
        return results

    async def search(self, query: str) -> list[Memory] | None:
        raise NotImplementedError()
